
"""
    ############################################################
    Gene Function Prediction (``examples.gene_func_prediction``)
    ############################################################
    
    As a background reading before this example, we recommend user to read [Schietgat2010]_ and [Schachtner2008]_ where
    the authors study the use of decision tree based models for predicting the multiple gene functions and unsupervised 
    matrix factorization techniques to extract marker genes from gene expression profiles for classification into
    diagnostic categories, respectively. 
        
    This example from functional genomics deals with predicting gene functions. Two main characteristics of gene function 
    prediction task are:
    
        #. single gene can have multiple functions, 
        #. the functions are organized in a hierarchy, in particular in a hierarchy structered as a rooted tree -- MIPS
           Functional Catalogue. A gene related to some function is automatically related to all its ancestor 
           functions. Data set used in this example originates from S. cerevisiae and has annotations from the MIPS 
           Functional Catalogue. 
    
    The latter problem setting describes hierarchical multi-label classification (HMC).
    
    .. note:: The S. cerevisiae FunCat annotated data set used in this example is not included in the `datasets`. If you 
              wish to perform the gene function prediction experiments, start by downloading the data set. In particular
              D1 (FC) seq data set must be available for the example to run.  Download links are listed in the 
              ``datasets``. To run the example, uncompress the data and put it into corresponding data directory, namely 
              the extracted data set must exist in the ``S_cerevisiae_FC`` directory under ``datasets``. Once you have 
              the data installed, you are ready to start running the experiments.  
    
    Here is the outline of this gene function prediction task. 
    
        #. Reading S. cerevisiae sequence data, i. e. train, validation and test set. Reading meta data,  
           attributes' labels and class labels. Weights are used to distinguish direct and indirect class 
           memberships of genes in gene function classes according to FunCat annotations. 
        #. Preprocessing, i. e. normalizing data matrix of test data and data matrix of joined train and validation
           data. 
        #. Factorization of train data matrix. We used SNMF/L factorization algorithm for train data. 
        #. Factorization of test data matrix. We used SNMF/L factorization algorithm for train data.
        #. Application of rules for class assignments. Three rules can be used, average correlation and maximal 
           correlation, as in [Schachtner2008]_ and threshold maximal correlation. All class assignments rules
           are generalized to meet the hierarchy constraint imposed by the rooted tree structure of MIPS Functional 
           Catalogue. 
        #. Precision-recall (PR) evaluation measures. 
    
    To run the example simply type::
        
        python gene_func_prediction.py
        
    or call the module's function::
    
        import nimfa.examples
        nimfa.examples.gene_func_prediction.run()
        
    .. note:: This example uses ``matplotlib`` library for producing visual interpretation.
"""

from os.path import dirname, abspath
from os.path import join
from warnings import warn

import numpy as np

import nimfa

try:
    import matplotlib.pylab as plb
except ImportError as exc:
    warn("Matplotlib must be installed to run Gene Function prediction example.")


def run():
    """
    Run the gene function prediction example on the S. cerevisiae sequence data set (D1 FC seq).
    
    The methodology is as follows:
        #. Reading S. cerevisiae sequence data, i. e. train, validation and test set. Reading meta data,  
           attributes' labels and class labels.
        #. Preprocessing, i. e. normalizing data matrix of test data and data matrix of joined train and validation
           data. 
        #. Factorization of train data matrix. We used SNMF/L factorization algorithm for train data. 
        #. Factorization of test data matrix. We used SNMF/L factorization algorithm for train data.
        #. Application of rules for class assignments. Three rules can be used, average correlation and maximal 
           correlation, as in [Schachtner2008]_ and threshold maximal correlation. All class assignments rules
           are generalized to meet the hierarchy constraint imposed by the rooted tree structure of MIPS Functional 
           Catalogue. 
        #. PR evaluation measures. 
    """
    tv_data, test_data, idx2attr, idx2class = read()
    tv_data = preprocess(tv_data)
    test_data = preprocess(test_data)
    tv_data = factorize(tv_data)
    test_data = factorize(test_data)
    corrs = compute_correlations(tv_data, test_data)
    for method in 0.5 * np.random.random_sample(50) + 1.:
        func2gene = assign_labels(corrs, tv_data, idx2class, method=method)
        plot(func2gene, test_data, idx2class)


def read():
    """
    Read S. cerevisiae FunCat annotated sequence data set (D1 FC seq).
    
    Return attributes' values and class information of the test data set and joined train and validation data set. Additional mapping functions 
    are returned mapping attributes' names and classes' names to indices. 
    """
    print(" Reading S. cerevisiae FunCat annotated sequence data set (D1 FC seq)")
    dir = join(dirname(dirname(abspath(__file__))), 'datasets', 'S_cerevisiae_FC',  'seq_yeast_FUN')
    train_data = join(dir, 'seq_yeast_FUN.train.arff')
    valid_data = join(dir, 'seq_yeast_FUN.valid.arff')
    test_data = join(dir, 'seq_yeast_FUN.test.arff')
    print(" Reading S. cerevisiae FunCat annotated sequence (D1 FC seq) TRAIN set")
    train, idx2attr, idx2class = transform_data(
        train_data, include_meta=True)
    print(" Reading S. cerevisiae FunCat annotated sequence (D1 FC seq) VALIDATION set")
    valid = transform_data(valid_data)
    print(" Reading S. cerevisiae FunCat annotated sequence (D1 FC seq) TEST set")
    test = transform_data(test_data)
    print(" Joining S. cerevisiae FunCat annotated sequence (D1 FC seq) TEST and VALIDATION set")
    tv_data = _join(train, valid)
    return tv_data, test, idx2attr, idx2class


def transform_data(path, include_meta=False):
    """
    Read data in the ARFF format and transform it to suitable matrix for factorization process. For each feature update direct and indirect 
    class information exploiting properties of Functional Catalogue hierarchy. 
    
    Return attributes' values and class information. If :param:`include_meta` is specified additional mapping functions are provided with 
    mapping from indices to attributes' names and indices to classes' names.  
    
    :param path: Path of directory with sequence data set (D1 FC seq).
    :type path: `str`
    :param include_meta: Specify if the header of the ARFF file should be skipped. The header of the ARFF file 
                               contains the name of the relation, a list of the attributes and their types. Default
                               value is False.  
    :type include_meta: `bool`
    """
    class2idx = {}
    attr2idx = {}

    idx_attr = 0
    idx_class = 0
    idx = 0
    feature = 0
    used_idx = set()
    section = 'h'

    for line in open(path):
        if section == 'h':
            tokens = line.strip().split()
            line_type = tokens[0] if tokens else None
            if line_type == "@ATTRIBUTE":
                if tokens[2] in ["numeric"]:
                    attr2idx[tokens[1]] = idx_attr
                    idx_attr += 1
                    used_idx.add(idx)
                if tokens[1] in ["class"] and tokens[2] in ["hierarchical", "classes"]:
                    class2idx = _reverse(
                        dict(list(enumerate((tokens[3] if tokens[3] != '%' else tokens[5]).split(",")))))
                    idx_class = idx
                idx += 1
            if line_type == "@DATA":
                section = 'd'
                idxs = set(range(idx)).intersection(used_idx)
                attr_data = np.mat(np.zeros((1e4, len(attr2idx))))
                class_data = np.mat(np.zeros((1e4, len(class2idx))))
        elif section == 'd':
            d, _, comment = line.strip().partition("%")
            values = d.split(",")
            # update class information for current feature
            class_var = list(map(str.strip, values[idx_class].split("@")))
            for cl in class_var:
                # update direct class information
                class_data[feature, class2idx[cl]] += 10.
                # update indirect class information through FunCat hierarchy
                cl_a = cl.split("/")
                cl = "/".join(cl_a[:3] + ['0'])
                if cl in class2idx:
                    class_data[feature, class2idx[cl]] += 3.
                cl = "/".join(cl_a[:2] + ['0', '0'])
                if cl in class2idx:
                    class_data[feature, class2idx[cl]] += 2.
                cl = "/".join(cl_a[:1] + ['0', '0', '0'])
                if cl in class2idx:
                    class_data[feature, class2idx[cl]] += 1.
            # update attribute values information for current feature
            i = 0
            for idx in idxs:
                attr_data[feature, i] = abs(
                    float(values[idx] if values[idx] != '?' else 0.))
                i += 1
            feature += 1
    if include_meta:
        data = ({'feat': feature, 'attr': attr_data[:feature, :], 'class': class_data[:feature, :]},
                _reverse(attr2idx), _reverse(class2idx))
    else:
        data = {'feat': feature, 'attr': attr_data[:feature, :], 'class': class_data[:feature, :]}
    return data


def _join(train, valid):
    """
    Join test and validation data of the S. cerevisiae FunCat annotated sequence data set (D1 FC seq). 
    
    Return joined test and validation attributes' values and class information.
     
    :param train: Attributes' values and class information of the train data set. 
    :type train: `numpy.matrix`
    :param valid: Attributes' values and class information of the validation data set.
    :type valid: `numpy.matrix`
    """
    n_train = train['feat']
    n_valid = valid['feat']
    return {'feat': n_train + n_valid,
            'attr': np.vstack((train['attr'][:n_train, :], valid['attr'][:n_valid, :])),
            'class': np.vstack((train['class'][:n_train, :], valid['class'][:n_valid, :]))}


def _reverse(object2idx):
    """
    Reverse 1-to-1 mapping function.
    
    Return reversed mapping.
    
    :param object2idx: Mapping of objects to indices or vice verse.
    :type object2idx: `dict`
    :rtype: `dict`
    """
    return dict(list(zip(list(object2idx.values()), list(object2idx.keys()))))


def preprocess(data):
    """
    Preprocess S.cerevisiae FunCat annotated sequence data set (D1 FC seq). Preprocessing step includes data 
    normalization.
    
    Return preprocessed data. 
    
    :param data: Transformed data set containing attributes' values, class information and possibly additional meta information.  
    :type data: `tuple`
    """
    print("Data preprocessing")
    data['attr'] = (data['attr'] - data['attr'].min() + np.finfo(
        data['attr'].dtype).eps) / (data['attr'].max() - data['attr'].min())
    return data


def factorize(data):
    """
    Perform factorization on S. cerevisiae FunCat annotated sequence data set (D1 FC seq).
    
    Return factorized data, this is matrix factors as result of factorization (basis and mixture matrix). 
    
    :param data: Transformed data set containing attributes' values, class information and possibly additional meta information.  
    :type data: `tuple`
    """
    V = data['attr']
    snmf = nimfa.Snmf(V, seed="random_vcol", rank=40, max_iter=5, version="l", eta=1., beta=1e-4,
                     i_conv=10, w_min_change=0)
    print("Algorithm: %s\nInitialization: %s\nRank: %d" % (snmf, snmf.seed, snmf.rank))
    fit = snmf()
    sparse_w, sparse_h = fit.fit.sparseness()
    print("""Stats:
            - iterations: %d
            - KL Divergence: %5.3f
            - Euclidean distance: %5.3f
            - Sparseness basis: %5.3f, mixture: %5.3f""" % (fit.fit.n_iter, fit.distance(), fit.distance(metric='euclidean'), sparse_w, sparse_h))
    data['W'] = fit.basis()
    data['H'] = fit.coef()
    return data


def compute_correlations(train, test):
    """
    Estimate correlation coefficients between profiles of train basis matrix and profiles of test basis matrix. 
    
    Return the estimated correlation coefficients of the features (variables).  
    
    :param train: Factorization matrix factors of train data set. 
    :type train: `dict`
    :param test: Factorization matrix factors of test data set. 
    :type test: `dict`
    :rtype: `numpy.matrix`
    """
    print("Estimating correlation coefficients")
    corrs = np.corrcoef(train['W'], test['W'])
    # alternative, it is time consuming - can be used for partial evaluation
    """corrs = {}
    for i in xrange(test['W'].shape[0]):
        corrs.setdefault(i, np.mat(np.zeros((train['W'].shape[0], 1))))
        for j in xrange(train['W'].shape[0]):
            corrs[i][j, 0] = _corr(test['W'][i, :], train['W'][j, :])"""
    return np.mat(corrs)


def _corr(x, y):
    """
    Compute Pearson's correlation coefficient of x and y. Numerically stable algebraically equivalent equation for 
    coefficient computation is used. 
    
    Return correlation coefficient between x and y which is by definition in [-1, 1].
    
    :param x: Random variable.
    :type x: `numpy.matrix`
    :param y: Random variable.
    :type y: `numpy.matrix`
    :rtype: `float`
    """
    xc = (x - x.mean()) / x.std(ddof=1)
    yc = (y - y.mean()) / y.std(ddof=1)
    return 1. / (x.size - 1) * np.multiply(xc, yc).sum()


def assign_labels(corrs, train, idx2class, method=0.):
    """
    Apply rules for class assignments. In [Schachtner2008]_ two rules are proposed, average correlation and maximal 
    correlation. Here, both the rules are implemented and can be specified through :param:`method``parameter. In addition to 
    these the threshold maximal correlation rule is possible as well. Class assignments rules are generalized to 
    multi-label classification incorporating hierarchy constraints. 
    
    User can specify the usage of one of the following rules:
        #. average correlation,
        #. maximal correlation,
        #. threshold maximal correlation.
    
    Though any method based on similarity measures can be used, we estimate correlation coefficients. Let w be the
    gene profile of test basis matrix for which we want to predict gene functions. For each class C a separate 
    index set A of indices is created, where A encompasses all indices m, for which m-th profile of train basis 
    matrix has label C. Index set B contains all remaining indices. Now, the average correlation coefficient between w
    and elements of A is computed, similarly average correlation coefficient between w and elements of B. Finally, 
    w is assigned label C if the former correlation over the respective index set is greater than the 
    latter correlation.
    
    .. note:: Described rule assigns the class label according to an average correlation of test vector with all
              vectors belonging to one or the other index set. Minor modification of this rule is to assign the class
              label according to the maximal correlation occurring between the test vector and the members of each
              index set. 
             
    .. note:: As noted before the main problem of this example is the HMC (hierarchical multi-label classification) 
              setting. Therefore we generalized the concepts from articles describing the use of factorization
              for binary classification problems to multi-label classification. Additionally, we use the weights
              for class memberships to incorporate hierarchical structure of MIPS MIPS Functional
              Catalogue.
    
    Return mapping of gene functions to genes.  
    
    :param corrs: Estimated correlation coefficients between profiles of train basis matrix and profiles of test 
                  basis matrix. 
    :type corrs: `dict`
    :param train: Class information of train data set. 
    :type train: `dict`
    :param idx2class: Mapping between classes' indices and classes' labels. 
    :type idx2class: `dict`
    :param method: Type of rule for class assignments. Possible are average correlation, maximal correlation by 
                   specifying ``average`` or ``maximal`` respectively. In addition threshold maximal correlation is
                   supported. If threshold rule is desired, threshold is specified instead. By default 
                   threshold rule is applied. 
    :type method: `float` or `str`
    :rtype: `dict`
    """
    print("Assigning class labels - gene functions to genes")
    func2gene = {}
    n_train = train['feat']
    n_cl = len(idx2class)
    for cl_idx in range(n_cl):
        func2gene.setdefault(cl_idx, [])
    key = 0
    for test_idx in range(n_train, corrs.shape[0]):
        if method == "average":
            # weighted summation of correlations over respective index sets
            avg_corr_A = np.sum(
                np.multiply(np.tile(corrs[:n_train, test_idx], (1, n_cl)), train['class']), 0)
            avg_corr_B = np.sum(
                np.multiply(np.tile(corrs[:n_train, test_idx], (1, n_cl)), train['class'] != 0), 0)
            avg_corr_A = avg_corr_A / (np.sum(train['class'] != 0, 0) + 1)
            avg_corr_B = avg_corr_B / (np.sum(train['class'] == 0, 0) + 1)
            for cl_idx in range(n_cl):
                if (avg_corr_A[0, cl_idx] > avg_corr_B[0, cl_idx]):
                    func2gene[cl_idx].append(key)
        elif method == "maximal":
            max_corr_A = np.amax(
                np.multiply(np.tile(corrs[:n_train, test_idx], (1, n_cl)), train['class']), 0)
            max_corr_B = np.amax(
                np.multiply(np.tile(corrs[:n_train, test_idx], (1, n_cl)), train['class'] != 0), 0)
            for cl_idx in range(n_cl):
                if (max_corr_A[0, cl_idx] > max_corr_B[0, cl_idx]):
                    func2gene[cl_idx].append(key)
        elif isinstance(method, float):
            max_corr = np.amax(
                np.multiply(np.tile(corrs[:n_train, test_idx], (1, n_cl)), train['class']), 0)
            for cl_idx in range(n_cl):
                if (max_corr[0, cl_idx] >= method):
                    func2gene[cl_idx].append(key)
        else:
            raise ValueError("Unrecognized class assignment rule.")
        key += 1
        if key % 100 == 0:
            print(" %d/%d" % (key, corrs.shape[0] - n_train))
    return func2gene


def plot(func2gene, test, idx2class):
    """
    Report the performance with the precision-recall (PR) based evaluation measures. 
    
    Beside PR also ROC based evaluations have been used before to evaluate gene function prediction approaches. PR
    based better suits the characteristics of the common HMC task, in which many classes are infrequent with a small
    number of genes having particular function. That is for most classes the number of negative instances exceeds
    the number of positive instances. Therefore it is sometimes preferred to recognize the positive instances instead
    of correctly predicting the negative ones (i. e. gene does not have a particular function). That means that ROC
    curve might be less suited for the task as they reward a learner if it correctly predicts negative instances. 
    
    Return PR evaluations measures
    
    :param labels: Mapping of genes to their predicted gene functions. 
    :type labels: `dict`
    :param test: Class information of test data set. 
    :type test: `dict`
    :param idx2class: Mapping between classes' indices and classes' labels. 
    :type idx2class: `dict`
    :rtype: `tuple`
    """
    print("Computing PR evaluations measures")

    def tp(g_function):
        # number of true positives for g_function (correctly predicted positive
        # instances)
        return (test['class'][func2gene[g_function], g_function] != 0).sum()

    def fp(g_function):
        # number of false positives for g_function (positive predictions that
        # are incorrect)
        return (test['class'][func2gene[g_function], g_function] == 0).sum()

    def fn(g_function):
        # number of false negatives for g_function (positive instances that are
        # incorrectly predicted negative)
        n_pred = list(
            set(range(len(idx2class))).difference(func2gene[g_function]))
        return (test['class'][n_pred, g_function] != 0).sum()
    tp_sum = 0.
    fp_sum = 0.
    fn_sum = 0.
    for g_function in idx2class:
        tp_sum += tp(g_function)
        fp_sum += fp(g_function)
        fn_sum += fn(g_function)
    avg_precision = tp_sum / (tp_sum + fp_sum)
    avg_recall = tp_sum / (tp_sum + fn_sum)
    print("Average precision over all gene functions: %5.3f" % avg_precision)
    print("Average recall over all gene functions: %5.3f" % avg_recall)
    return avg_precision, avg_recall


if __name__ == "__main__":
    """Run the gene function prediction example."""
    run()



import numpy as np

import nimfa

V = np.random.rand(40, 100)
lsnmf = nimfa.Lsnmf(V, seed="random_vcol", rank=10, max_iter=12, sub_iter=10,
                    inner_sub_iter=10, beta=0.1)
lsnmf_fit = lsnmf()




"""
    ######################################
    CBCL Images (``examples.cbcl_images``)
    ######################################
    
    In this example of image processing we consider the problem demonstrated in [Lee1999]_.
    
    We used the CBCL face images database consisting of 2429 face images of size 19 x 19. The facial images 
    consist of frontal views hand aligned in a 19 x 19 grid. Each face image is preprocessed. For each image, 
    the greyscale intensities are first linearly scaled, so that the pixel mean and standard deviation are
    equal to 0.25, and then clipped to the range [0, 1].  
    
    .. note:: The CBCL face images database used in this example is not included in the `datasets`. If you wish to
              perform the CBCL data experiments, start by downloading the images.  Download links are listed in the 
              ``datasets``. To run the example, uncompress the data and put it into corresponding data directory, namely 
              the extracted CBCL data set must exist in the ``CBCL_faces`` directory under ``datasets``. Once you have 
              the data installed, you are ready to start running the experiments. 
      
    We experimented with the following factorization algorithms to learn the basis images from the CBCL database: 
    Standard NMF - Euclidean, LSNMF, SNMF/R and SNMF/L. The number of bases is 49. Random Vcol algorithm is used for factorization
    initialization. The algorithms mostly converge after less than 50 iterations. 
     
    Unlike vector quantization and principal components analysis ([Lee1999]_), these algorithms learn a parts-based representations of 
    faces and some also spatially localized representations depending on different types of constraints on basis and mixture matrix. 
    Following are 7 x 7 montages of learned basis images by different factorization algorithms. 
      
    .. figure:: /images/cbcl_faces_50_iters_LSNMF.png
       :scale: 90 %
       :alt: Basis images of LSNMF obtained after 50 iterations on original CBCL face images. 
       :align: center
       
       Basis images of LSNMF obtained after 50 iterations on original CBCL face images. The bases trained by LSNMF are additive
       but not spatially localized for representation of faces. 10 subiterations and 10 inner subiterations are performed
       (these are LSNMF specific parameters). 
       
       
    .. figure:: /images/cbcl_faces_50_iters_NMF.png
       :scale: 90 %
       :alt: Basis images of NMF obtained after 50 iterations on original CBCL face images. 
       :align: center
       
       Basis images of NMF obtained after 50 iterations on original CBCL face images. The images show that
       the bases trained by NMF are additive but not spatially localized for representation of faces. 
       
        
    .. figure:: /images/cbcl_faces_10_iters_SNMF_L.png
       :scale: 90 %
       :alt: Basis images of LSNMF obtained after 10 iterations on original CBCL face images. 
       :align: center
       
       Basis images of SNMF/L obtained after 10 iterations on original CBCL face images. The
       bases trained from LSNMF/L are both additive and spatially localized for representing faces. LSNMF/L imposes
       sparseness constraints on basis matrix, whereas LSNMF/R imposes sparseness on mixture matrix. Therefore obtained basis images
       are very sparse as it can be shown in the figure. The Euclidean distance of SNMF/L estimate from target matrix is 1827.66.  
       
       
    .. figure:: /images/cbcl_faces_10_iters_SNMF_R.png
       :scale: 90 %
       :alt: Basis images of SNMF/R obtained after 10 iterations on original CBCL face images. 
       :align: center
       
       Basis images of SNMF/R obtained after 10 iterations on original CBCL face images. The images show that
       the bases trained by NMF are additive but not spatially localized for representation of faces. The Euclidean
       distance of SNMF/R estimate from target matrix is 3948.149. 
       
          
    To run the example simply type::
        
        python cbcl_images.py
        
    or call the module's function::
    
        import nimfa.examples
        nimfa.examples.cbcl_images.run()
        
    .. note:: This example uses ``matplotlib`` library for producing visual interpretation of basis vectors. It uses PIL 
              library for displaying face images. 
    
"""
from os.path import dirname, abspath
from os.path import join
from warnings import warn

import numpy as np

import nimfa

try:
    from matplotlib.pyplot import savefig, imshow, set_cmap
except ImportError as exc:
    warn("Matplotlib must be installed to run CBCL images example.")

try:
    from PIL.Image import open, fromarray, new
    from PIL.ImageOps import expand
except ImportError as exc:
    warn("PIL must be installed to run CBCL images example.")


def run():
    """Run LSNMF on CBCL faces data set."""
    V = read()
    V = preprocess(V)
    W, _ = factorize(V)
    plot(W)


def factorize(V):
    """
    Perform LSNMF factorization on the CBCL faces data matrix. 
    
    Return basis and mixture matrices of the fitted factorization model. 
    
    :param V: The CBCL faces data matrix. 
    :type V: `numpy.matrix`
    """
    lsnmf = nimfa.Lsnmf(V, seed="random_vcol", rank=49, max_iter=50, sub_iter=10,
                        inner_sub_iter=10, beta=0.1, min_residuals=1e-8)
    print("Algorithm: %s\nInitialization: %s\nRank: %d" % (lsnmf, lsnmf.seed, lsnmf.rank))
    fit = lsnmf()
    sparse_w, sparse_h = fit.fit.sparseness()
    print("""Stats:
            - iterations: %d
            - final projected gradients norm: %5.3f
            - Euclidean distance: %5.3f 
            - Sparseness basis: %5.3f, mixture: %5.3f""" % (fit.fit.n_iter,
                                                            fit.distance(),
                                                            fit.distance(metric='euclidean'),
                                                            sparse_w, sparse_h))
    return fit.basis(), fit.coef()


def read():
    """
    Read face image data from the CBCL database. The matrix's shape is 361 (pixels) x 2429 (faces). 
    
    Step through each subject and each image. Images' sizes are not reduced.  
    
    Return the CBCL faces data matrix. 
    """
    print("Reading CBCL faces database")
    dir = join(dirname(dirname(abspath(__file__))), 'datasets', 'CBCL_faces', 'face')
    V = np.zeros((19 * 19, 2429))
    for image in range(2429):
        im = open(join(dir, "face0%s.pgm" % str(image + 1).zfill(4)))
        V[:, image] = np.asarray(im).flatten()
    return V


def preprocess(V):
    """
    Preprocess CBCL faces data matrix as Lee and Seung.
    
    Return normalized and preprocessed data matrix. 
    
    :param V: The CBCL faces data matrix. 
    :type V: `numpy.matrix`
    """
    print("Data preprocessing")
    V = (V - V.mean()) / np.sqrt(np.multiply(V, V).mean())
    V = np.maximum(np.minimum((V + 0.25) * 0.25, 1), 0)
    return V


def plot(W):
    """
    Plot basis vectors.
    
    :param W: Basis matrix of the fitted factorization model.
    :type W: `numpy.matrix`
    """
    set_cmap('gray')
    blank = new("L", (133 + 6, 133 + 6))
    for i in range(7):
        for j in range(7):
            basis = np.array(W[:, 7 * i + j])[:, 0].reshape((19, 19))
            basis = basis / np.max(basis) * 255
            basis = 255 - basis
            ima = fromarray(basis)
            ima = ima.rotate(180)
            expand(ima, border=1, fill='black')
            blank.paste(ima.copy(), (j * 19 + j, i * 19 + i))
    imshow(blank)
    savefig("cbcl_faces.png")


if __name__ == "__main__":
    """Run the CBCL faces example."""
    run()




"""
    ##############################################
    Medulloblastoma (``examples.medulloblastoma``)
    ##############################################

    This module demonstrates the ability of NMF to recover meaningful biological information from childhood 
    brain tumors microarray data. 
    
    Medulloblastoma data set is used in this example. The pathogenesis of these childhood brain tumors is not well 
    understood but is accepted that there are two known histological subclasses; classic (C) and desmoplastic (D). 
    These subclasses can be clearly seen under microscope.   
    
    .. note:: Medulloblastoma data set used in this example is included in the `datasets` and does not need to be
          downloaded. However, download links are listed in the ``datasets``. To run the example, the data set
          must exist in the ``Medulloblastoma`` directory under `datasets`. 
    
    This example is inspired by [Brunet2004]_. In [Brunet2004]_ authors applied NMF to the medulloblastoma data set and managed to expose a
    separate desmoplastic (D) class. In [Brunet2004]_ authors also applied SOM and HC to these data but were unable to find a distinct
    desmoplastic class. Using HC desmoplastic samples were scattered among leaves and there was no level of the tree
    where they could split the branches to expose a clear desmoplastic cluster. They applied SOM by using two to eight 
    centroids but did not recover distinct desmoplastic class as well. 
    
    .. figure:: /images/medulloblastoma_consensus2.png
       :scale: 60 %
       :alt: Consensus matrix generated for rank, rank = 2.
       :align: center 
       
       Reordered consensus matrix generated for rank, rank = 2. Reordered consensus matrix averages 50 connectivity 
       matrices computed at rank = 2, 3 for the medulloblastoma data set consisting of 25 classic and 9 desmoplastic
       medulloblastoma tumors. Consensus matrix is reordered with HC by using distances derived from consensus clustering 
       matrix entries, coloured from 0 (deep blue, samples are never in the same cluster) to 1 (dark red, samples are 
       always in the same cluster).   
       
    .. figure:: /images/medulloblastoma_consensus3.png
       :scale: 60 %
       :alt: Consensus matrix generated for rank, rank = 3. 
       :align: center
       
       Reordered consensus matrix generated for rank, rank = 3.
       
       
    .. table:: Standard NMF Class assignments results obtained with this example for rank = 2, rank = 3 and rank = 5.  

       ====================  ========== ========== ========== ==========
              Sample           Class     rank = 2   rank = 3   rank = 5 
       ====================  ========== ========== ========== ==========
        Brain_MD_7                C        0            1        3
        Brain_MD_59               C        1            0        2
        Brain_MD_20               C        1            1        3
        Brain_MD_21               C        1            1        3
        Brain_MD_50               C        1            1        4
        Brain_MD_49               C        0            2        3
        Brain_MD_45               C        1            1        3
        Brain_MD_43               C        1            1        3
        Brain_MD_8                C        1            1        3
        Brain_MD_42               C        0            2        4
        Brain_MD_1                C        0            2        3
        Brain_MD_4                C        0            2        3 
        Brain_MD_55               C        0            2        3
        Brain_MD_41               C        1            1        2
        Brain_MD_37               C        1            0        3
        Brain_MD_3                C        1            2        3
        Brain_MD_34               C        1            2        4
        Brain_MD_29               C        1            1        2
        Brain_MD_13               C        0            1        2
        Brain_MD_24               C        0            1        3
        Brain_MD_65               C        1            0        2
        Brain_MD_5                C        1            0        1
        Brain_MD_66               C        1            0        1
        Brain_MD_67               C        1            0        3
        Brain_MD_58               C        0            2        3
        Brain_MD_53               D        0            2        4
        Brain_MD_56               D        0            2        4
        Brain_MD_16               D        0            2        4
        Brain_MD_40               D        0            1        0
        Brain_MD_35               D        0            2        4
        Brain_MD_30               D        0            2        4
        Brain_MD_23               D        0            2        4
        Brain_MD_28               D        1            2        1
        Brain_MD_60               D        1            0        0
       ====================  ========== ========== ========== ==========   
    
    To run the example simply type::
        
        python medulloblastoma.py
        
    or call the module's function::
    
        import nimfa.examples
        nimfa.examples.medulloblastoma.run()
        
    .. note:: This example uses ``matplotlib`` library for producing a heatmap of a consensus matrix.
"""

from os.path import dirname, abspath
from os.path import join
from warnings import warn

from scipy.cluster.hierarchy import linkage, leaves_list
from scipy.spatial.distance import squareform
import numpy as np

import nimfa

try:
    from matplotlib.pyplot import savefig, imshow, set_cmap
except ImportError as exc:
    warn("Matplotlib must be installed to run Medulloblastoma example.")


def run():
    """Run Standard NMF on medulloblastoma data set. """
    V = read()
    for rank in range(2, 4):
        run_one(V, rank)


def run_one(V, rank):
    """
    Run standard NMF on medulloblastoma data set. 50 runs of Standard NMF are performed and obtained consensus matrix
    averages all 50 connectivity matrices.  
    
    :param V: Target matrix with gene expression data.
    :type V: `numpy.ndarray`
    :param rank: Factorization rank.
    :type rank: `int`
    """
    print("================= Rank = %d =================" % rank)
    consensus = np.zeros((V.shape[1], V.shape[1]))
    for i in range(50):
        nmf = nimfa.Nmf(V, rank=rank, seed="random_vcol", max_iter=200, update='euclidean',
                         objective='conn', conn_change=40)
        fit = nmf()
        print("Algorithm: %s\nInitialization: %s\nRank: %d" % (nmf, nmf.seed, nmf.rank))
        consensus += fit.fit.connectivity()
    consensus /= 50.
    p_consensus = reorder(consensus)
    plot(p_consensus, rank)


def plot(C, rank):
    """
    Plot reordered consensus matrix.
    
    :param C: Reordered consensus matrix.
    :type C: `numpy.ndarray`
    :param rank: Factorization rank.
    :type rank: `int`
    """
    set_cmap("RdBu_r")
    imshow(C)
    savefig("medulloblastoma_consensus_%s.png" % rank)


def reorder(C):
    """
    Reorder consensus matrix.
    
    :param C: Consensus matrix.
    :type C: `numpy.ndarray`
    """
    Y = 1 - C
    Z = linkage(squareform(Y), method='average')
    ivl = leaves_list(Z)
    ivl = ivl[::-1]
    return C[:, ivl][ivl, :]


def read(normalize=False):
    """
    Read the medulloblastoma gene expression data. The matrix's shape is 5893 (genes) x 34 (samples). 
    It contains only positive data.
    
    Return the gene expression data matrix. 
    """
    fname = join(dirname(dirname(abspath(__file__))), 'datasets', 'Medulloblastoma',  'Medulloblastoma_data.txt')
    V = np.loadtxt(fname)
    if normalize:
        V = (V - V.min()) / (V.max() - V.min())
    return V


if __name__ == "__main__":
    """Run the medulloblastoma example."""
    run()



"""
    #########################
    Linalg (``utils.linalg``)
    #########################
    
    Linear algebra helper routines and wrapper functions for handling sparse
    matrices and dense matrices representation.
"""

import sys
import copy
import numpy as np
import scipy
import scipy.sparse as sp
import scipy.sparse.linalg as sla
import numpy.linalg as nla
from operator import mul, eq, ne, add, ge, le, itemgetter
from operator import truediv as div

from math import sqrt, log, isnan, ceil
from scipy.cluster.hierarchy import linkage, cophenet
from scipy.special import erfc, erfcinv
import warnings

#
# Wrapper functions for handling sparse matrices and dense matrices representation.
###    scipy.sparse, numpy.matrix
#


def diff(X):
    """
    Compute differences between adjacent elements of X.

    :param X: Vector for which consecutive differences are computed.
    :type X: :class:`numpy.matrix`
    """
    assert 1 in X.shape, "sX should be a vector."
    assert not sp.isspmatrix(X), "X is sparse matrix."
    X = X.flatten()
    return [X[0, j + 1] - X[0, j] for j in range(X.shape[1] - 1)]


def sub2ind(shape, row_sub, col_sub):
    """
    Return the linear index equivalents to the row and column subscripts for
    given matrix shape.

    :param shape: Preferred matrix shape for subscripts conversion.
    :type shape: `tuple`
    :param row_sub: Row subscripts.
    :type row_sub: `list`
    :param col_sub: Column subscripts.
    :type col_sub: `list`
    """
    assert len(row_sub) == len(
        col_sub), "Row and column subscripts do not match."
    res = [j * shape[0] + i for i, j in zip(row_sub, col_sub)]
    return res


def trace(X):
    """
    Return trace of sparse or dense square matrix X.

    :param X: Target matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil,
    dia or :class:`numpy.matrix`
    """
    assert X.shape[0] == X.shape[1], "X should be square matrix."
    if sp.isspmatrix(X):
        return sum(X[i, i] for i in range(X.shape[0]))
    else:
        return np.trace(np.mat(X))


def any(X, axis=None):
    """
    Test whether any element along a given axis of sparse or dense matrix X is nonzero.

    :param X: Target matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil,
    dia or :class:`numpy.matrix`
    :param axis: Specified axis along which nonzero test is performed.
    If :param:`axis` not specified, whole matrix is considered.
    :type axis: `int`
    """
    if sp.isspmatrix(X):
        X = X.tocsr()
        assert axis == 0 or axis == 1 or axis is None, "Incorrect axis number."
        if axis is None:
            return len(X.data) != X.shape[0] * X.shape[1]
        res = [0 for _ in range(X.shape[1 - axis])]

        def _caxis(now, row, col):
            res[col] += 1

        def _raxis(now, row, col):
            res[row] += 1
        check = _caxis if axis == 0 else _raxis
        now = 0
        for row in range(X.shape[0]):
            upto = X.indptr[row + 1]
            while now < upto:
                col = X.indices[now]
                check(now, row, col)
                now += 1
        sol = [x != 0 for x in res]
        return np.mat(sol) if axis == 0 else np.mat(sol).T
    else:
        return X.any(axis)


def all(X, axis=None):
    """
    Test whether all elements along a given axis of sparse or dense matrix
    :param:`X` are nonzero.

    :param X: Target matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param axis: Specified axis along which nonzero test is performed.
    If :param:`axis` not specified, whole matrix is considered.
    :type axis: `int`
    """
    if sp.isspmatrix(X):
        X = X.tocsr()
        assert axis == 0 or axis == 1 or axis is None, "Incorrect axis number."
        if axis is None:
            return len(X.data) == X.shape[0] * X.shape[1]
        res = [0 for _ in range(X.shape[1 - axis])]

        def _caxis(now, row, col):
            res[col] += 1

        def _raxis(now, row, col):
            res[row] += 1
        check = _caxis if axis == 0 else _raxis
        now = 0
        for row in range(X.shape[0]):
            upto = X.indptr[row + 1]
            while now < upto:
                col = X.indices[now]
                check(now, row, col)
                now += 1
        sol = [x == X.shape[0] if axis == 0 else x == X.shape[1] for x in res]
        return np.mat(sol) if axis == 0 else np.mat(sol).T
    else:
        return X.all(axis)


def find(X):
    """
    Return all nonzero elements indices (linear indices) of sparse or dense
    matrix :param:`X`. It is Matlab notation.

    :param X: Target matrix.
    type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    """
    if sp.isspmatrix(X):
        X = X.tocsr()
        res = []
        now = 0
        for row in range(X.shape[0]):
            upto = X.indptr[row + 1]
            while now < upto:
                col = X.indices[now]
                if X.data[now]:
                    res.append(col * X.shape[0] + row)
                now += 1
        return res
    else:
        return [j * X.shape[0] + i for i in range(X.shape[0]) for j in range(X.shape[1]) if X[i, j]]


def negative(X):
    """
    Check if :param:`X` contains negative elements.

    :param X: Target matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil,
    dia or :class:`numpy.matrix`
    """
    if sp.isspmatrix(X):
        if any(X.data < 0):
            return True
    else:
        if any(np.asmatrix(X) < 0):
            return True


def sort(X):
    """
    Return sorted elements of :param:`X` and array of corresponding
    sorted indices.

    :param X: Target vector.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil,
    dia or :class:`numpy.matrix`
    """
    assert 1 in X.shape, "X should be vector."
    X = X.flatten().tolist()[0]
    return sorted(X), sorted(list(range(len(X))), key=X.__getitem__)


def std(X, axis=None, ddof=0):
    """
    Compute the standard deviation along the specified :param:`axis` of
    matrix :param:`X`.

    :param X: Target matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil,
    dia or :class:`numpy.matrix`
    :param axis: Axis along which deviation is computed. If not specified,
    whole matrix :param:`X` is considered.
    :type axis: `int`
    :param ddof: Means delta degrees of freedom. The divisor used in
    computation is N - :param:`ddof`, where N represents the
    number of elements. Default is 0.
    :type ddof: `float`
    """
    assert len(X.shape) == 2, "Input matrix X should be 2-D."
    assert axis == 0 or axis == 1 or axis is None, "Incorrect axis number."
    if sp.isspmatrix(X):
        if axis is None:
            mean = X.mean()
            no = X.shape[0] * X.shape[1]
            return sqrt(1. / (no - ddof) * sum((x - mean) ** 2 for x in X.data) + (no - len(X.data) * mean ** 2))
        if axis == 0:
            return np.mat([np.std(X[:, i].toarray(), axis, ddof) for i in range(X.shape[1])])
        if axis == 1:
            return np.mat([np.std(X[i, :].toarray(), axis, ddof) for i in range(X.shape[0])]).T
    else:
        return np.std(X, axis=axis, ddof=ddof)


def argmax(X, axis=None):
    """
    Return tuple (values, indices) of the maximum entries of matrix
    :param:`X` along axis :param:`axis`. Row major order.

    :param X: Target matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil,
    dia or :class:`numpy.matrix`
    :param axis: Specify axis along which to operate. If not specified,
    whole matrix :param:`X` is considered.
    :type axis: `int`
    """
    if sp.isspmatrix(X):
        X = X.tocsr()
        assert axis == 0 or axis == 1 or axis is None, "Incorrect axis number."
        res = [[float('-inf'), 0]
               for _ in range(X.shape[1 - axis])] if axis is not None else [float('-inf'), 0]

        def _caxis(row, col):
            if X[row, col] > res[col][0]:
                res[col] = (X[row, col], row)

        def _raxis(row, col):
            if X[row, col] > res[row][0]:
                res[row] = (X[row, col], col)

        def _naxis(row, col):
            if X[row, col] > res[0]:
                res[0] = X[row, col]
                res[1] = row * X.shape[0] + col
        check = _caxis if axis == 0 else _raxis if axis == 1 else _naxis
        [check(row, col) for row in range(X.shape[0])
         for col in range(X.shape[1])]
        if axis is None:
            return res
        elif axis == 0:
            t = list(zip(*res))
            return list(t[0]), np.mat(t[1])
        else:
            t = list(zip(*res))
            return list(t[0]), np.mat(t[1]).T
    else:
        idxX = np.asmatrix(X).argmax(axis)
        if axis is None:
            eX = X[idxX // X.shape[1], idxX % X.shape[1]]
        elif axis == 0:
            eX = [X[idxX[0, idx], col]
                  for idx, col in zip(range(X.shape[1]), range(X.shape[1]))]
        else:
            eX = [X[row, idxX[idx, 0]]
                  for row, idx in zip(range(X.shape[0]), range(X.shape[0]))]
        return eX, idxX


def argmin(X, axis=None):
    """
    Return tuple (values, indices) of the minimum entries of matrix :param:`X`
    along axis :param:`axis`. Row major order.

    :param X: Target matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil,
    dia or :class:`numpy.matrix`
    :param axis: Specify axis along which to operate. If not specified,
    whole matrix :param:`X` is considered.
    :type axis: `int`
    """
    if sp.isspmatrix(X):
        X = X.tocsr()
        assert axis == 0 or axis == 1 or axis is None, "Incorrect axis number."
        res = [[float('inf'), 0]
               for _ in range(X.shape[1 - axis])] if axis is not None else [float('inf'), 0]

        def _caxis(row, col):
            if X[row, col] < res[col][0]:
                res[col] = (X[row, col], row)

        def _raxis(row, col):
            if X[row, col] < res[row][0]:
                res[row] = (X[row, col], col)

        def _naxis(row, col):
            if X[row, col] < res[0]:
                res[0] = X[row, col]
                res[1] = row * X.shape[0] + col
        check = _caxis if axis == 0 else _raxis if axis == 1 else _naxis
        [check(row, col) for row in range(X.shape[0])
         for col in range(X.shape[1])]
        if axis is None:
            return res
        elif axis == 0:
            t = list(zip(*res))
            return list(t[0]), np.mat(t[1])
        else:
            t = list(zip(*res))
            return list(t[0]), np.mat(t[1]).T
    else:
        idxX = np.asmatrix(X).argmin(axis)
        if axis is None:
            eX = X[idxX // X.shape[1], idxX % X.shape[1]]
        elif axis == 0:
            eX = [X[idxX[0, idx], col]
                  for idx, col in zip(range(X.shape[1]), range(X.shape[1]))]
        else:
            eX = [X[row, idxX[idx, 0]]
                  for row, idx in zip(range(X.shape[0]), range(X.shape[0]))]
        return eX, idxX


def repmat(X, m, n):
    """
    Construct matrix consisting of an m-by-n tiling of copies of X.

    :param X: The input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil,
    dia or :class:`numpy.matrix`
    :param m,n: The number of repetitions of :param:`X` along each axis.
    :type m,n: `int`
    """
    if sp.isspmatrix(X):
        return sp.hstack([sp.vstack([X for _ in range(m)], format=X.format) for _ in range(n)], format=X.format)
    else:
        return np.tile(np.asmatrix(X), (m, n))


def inv_svd(X):
    """
    Compute matrix inversion using SVD.

    :param X: The input matrix.
    :type X: :class:`scipy.sparse` or :class:`numpy.matrix`
    """
    U, S, V = svd(X)
    if sp.isspmatrix(S):
        S_inv = _sop_spmatrix(S, op=lambda x: 1. / x)
    else:
        S_inv = np.diag(1. / np.diagonal(S))
    X_inv = dot(dot(V.T, S_inv), U.T)
    return X_inv


def svd(X):
    """
    Compute standard SVD on matrix X.

    :param X: The input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil,
    dia or :class:`numpy.matrix`
    """
    if sp.isspmatrix(X):
        if X.shape[0] <= X.shape[1]:
            U, S, V = _svd_left(X)
        else:
            U, S, V = _svd_right(X)
    else:
        U, S, V = nla.svd(np.mat(X), full_matrices=False)
        S = np.mat(np.diag(S))
    return U, S, V


def _svd_right(X):
    """
    Compute standard SVD on matrix X. Scipy.sparse.linalg.svd ARPACK does
    not allow computation of rank(X) SVD.

    :param X: The input sparse matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    """
    XXt = dot(X, X.T)
    if X.shape[0] > 1:
        if '0.8' in scipy.version.version:
            val, u_vec = sla.eigen_symmetric(XXt, k=X.shape[0] - 1)
        else:
            # In scipy 0.9.0 ARPACK interface has changed. eigen_symmetric
            # routine was renamed to eigsh
            # http://docs.scipy.org/doc/scipy/reference/release.0.9.0.html#scipy-sparse
            try:
                val, u_vec = sla.eigsh(XXt, k=X.shape[0] - 1)
            except sla.ArpackNoConvergence as err:
                # If eigenvalue iteration fails to converge, partially
                # converged results can be accessed
                val = err.eigenvalues
                u_vec = err.eigenvectors
    else:
        val, u_vec = nla.eigh(XXt.todense())
    # remove insignificant eigenvalues
    keep = np.where(val > 1e-7)[0]
    u_vec = u_vec[:, keep]
    val = val[keep]
    # sort eigen vectors (descending)
    idx = np.argsort(val)[::-1]
    val = val[idx]
    # construct U
    U = sp.csr_matrix(u_vec[:, idx])
    # compute S
    tmp_val = np.sqrt(val)
    tmp_l = len(idx)
    S = sp.spdiags(tmp_val, 0, m=tmp_l, n=tmp_l, format='csr')
    # compute V from inverse of S
    inv_S = sp.spdiags(1. / tmp_val, 0, m=tmp_l, n=tmp_l, format='csr')
    V = U.T * X
    V = inv_S * V
    return U, S, V


def _svd_left(X):
    """
    Compute standard SVD on matrix X. Scipy.sparse.linalg.svd ARPACK does
    not allow computation of rank(X) SVD.

    :param X: The input sparse matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    """
    XtX = dot(X.T, X)
    if X.shape[1] > 1:
        if '0.9' in scipy.version.version or '0.10' in scipy.version.version or '0.11' in scipy.version.version:
            # In scipy 0.9.0 ARPACK interface has changed. eigen_symmetric
            # routine was renamed to eigsh
            # http://docs.scipy.org/doc/scipy/reference/release.0.9.0.html#scipy-sparse
            try:
                val, v_vec = sla.eigsh(XtX, k=X.shape[1] - 1)
            except sla.ArpackNoConvergence as err:
                # If eigenvalue iteration fails to converge, partially
                # converged results can be accessed
                val = err.eigenvalues
                v_vec = err.eigenvectors
        else:
            val, v_vec = sla.eigen_symmetric(XtX, k=X.shape[1] - 1)
    else:
        val, v_vec = nla.eigh(XtX.todense())
    # remove insignificant eigenvalues
    keep = np.where(val > 1e-7)[0]
    v_vec = v_vec[:, keep]
    val = val[keep]
    # sort eigen vectors (descending)
    idx = np.argsort(val)[::-1]
    val = val[idx]
    # construct V
    V = sp.csr_matrix(v_vec[:, idx])
    # compute S
    tmp_val = np.sqrt(val)
    tmp_l = len(idx)
    S = sp.spdiags(tmp_val, 0, m=tmp_l, n=tmp_l, format='csr')
    # compute U from inverse of S
    inv_S = sp.spdiags(1. / tmp_val, 0, m=tmp_l, n=tmp_l, format='csr')
    U = X * V * inv_S
    V = V.T
    return U, S, V


def dot(X, Y):
    """
    Compute dot product of matrices :param:`X` and :param:`Y`.

    :param X: First input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param Y: Second input matrix.
    :type Y: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    """
    if sp.isspmatrix(X) and sp.isspmatrix(Y):
        return X * Y
    elif sp.isspmatrix(X) or sp.isspmatrix(Y):
        # avoid dense dot product with mixed factors
        return sp.csr_matrix(X) * sp.csr_matrix(Y)
    else:
        return np.asmatrix(X) * np.asmatrix(Y)


def multiply(X, Y):
    """
    Compute element-wise multiplication of matrices :param:`X` and :param:`Y`.

    :param X: First input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param Y: Second input matrix.
    :type Y: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    """
    if sp.isspmatrix(X) and sp.isspmatrix(Y):
        return X.multiply(Y)
    elif sp.isspmatrix(X) or sp.isspmatrix(Y):
        return _op_spmatrix(X, Y, np.multiply)
    else:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            return np.multiply(np.mat(X), np.mat(Y))


def power(X, s):
    """
    Compute matrix power of matrix :param:`X` for power :param:`s`.

    :param X: Input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param s: Power.
    :type s: `int`
    """
    if sp.isspmatrix(X):
        Y = X.tocsr()
        eps = np.finfo(Y.data.dtype).eps if not 'int' in str(
            Y.data.dtype) else 0
        return sp.csr_matrix((np.power(Y.data + eps, s), Y.indices, Y.indptr), Y.shape)
    else:
        eps = np.finfo(X.dtype).eps if not 'int' in str(X.dtype) else 0
        return np.power(X + eps, s)


def sop(X, s=None, op=None):
    """
    Compute scalar element wise operation of matrix :param:`X` and
    scalar :param:`s`.

    :param X: The input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param s: Input scalar. If not specified, element wise operation of input
    matrix is computed.
    :type s: `float`
    :param op: Operation to be performed.
    :type op: `func`
    """
    if sp.isspmatrix(X):
        return _sop_spmatrix(X, s, op)
    else:
        return _sop_matrix(X, s, op)


def _sop_spmatrix(X, s=None, op=None):
    """
    Compute sparse scalar element wise operation of matrix X and scalar :param:`s`.

    :param X: The input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    :param s: Input scalar. If not specified, element wise operation of input
    matrix is computed.
    :type s: `float`
    :param op: Operation to be performed.
    :type op: `func`
    """
    R = X.copy().tocsr()
    eps = np.finfo(R.dtype).eps if not 'int' in str(R.dtype) else 0
    now = 0
    for row in range(R.shape[0]):
        upto = R.indptr[row + 1]
        while now < upto:
            R.data[now] = op(R.data[now] + eps, s) if s is not None else op(
                R.data[now] + eps)
            now += 1
    return R


def _sop_matrix(X, s=None, op=None):
    """
    Compute scalar element wise operation of matrix :param:`X` and scalar :param:`s`.

    :param X: The input matrix.
    :type X: :class:`numpy.matrix`
    :param s: Input scalar. If not specified, element wise operation of input
    matrix is computed.
    :type s: `float`
    :param op: Operation to be performed.
    :type op: `func`
    """
    eps = np.finfo(X.dtype).eps if not 'int' in str(X.dtype) else 0
    return op(X + eps, s) if s is not None else op(X + eps)


def elop(X, Y, op):
    """
    Compute element-wise operation of matrix :param:`X` and matrix :param:`Y`.

    :param X: First input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param Y: Second input matrix.
    :type Y: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param op: Operation to be performed.
    :type op: `func`
    """
    try:
        zp1 = op(0, 1) if sp.isspmatrix(X) else op(1, 0)
        zp2 = op(0, 0)
        zp = zp1 != 0 or zp2 != 0
    except:
        zp = 0
    if sp.isspmatrix(X) or sp.isspmatrix(Y):
        return _op_spmatrix(X, Y, op) if not zp else _op_matrix(X, Y, op)
    else:
        try:
            X[X == 0] = np.finfo(X.dtype).eps
            Y[Y == 0] = np.finfo(Y.dtype).eps
        except ValueError:
            return op(np.mat(X), np.mat(Y))
        return op(np.mat(X), np.mat(Y))


def _op_spmatrix(X, Y, op):
    """
    Compute sparse element-wise operation for operations preserving zeros.

    :param X: First input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param Y: Second input matrix.
    :type Y: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param op: Operation to be performed.
    :type op: `func`
    """
    # distinction as op is not necessarily commutative
    return __op_spmatrix(X, Y, op) if sp.isspmatrix(X) else __op_spmatrix(Y, X, op)


def __op_spmatrix(X, Y, op):
    """
    Compute sparse element-wise operation for operations preserving zeros.

    :param X: First input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    :param Y: Second input matrix.
    :type Y: :class:`numpy.matrix`
    :param op: Operation to be performed.
    :type op: `func`
    """
    assert X.shape == Y.shape, "Matrices are not aligned."
    eps = np.finfo(Y.dtype).eps if not 'int' in str(Y.dtype) else 0
    Xx = X.tocsr()
    r, c = Xx.nonzero()
    R = op(Xx[r, c], Y[r, c] + eps)
    R = np.array(R)
    assert 1 in R.shape, "Data matrix in sparse should be rank-1."
    R = R[0, :] if R.shape[0] == 1 else R[:, 0]
    return sp.csr_matrix((R, Xx.indices, Xx.indptr), Xx.shape)


def _op_matrix(X, Y, op):
    """
    Compute sparse element-wise operation for operations not preserving zeros.

    :param X: First input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param Y: Second input matrix.
    :type Y: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param op: Operation to be performed.
    :type op: `func`
    """
    # operation is not necessarily commutative
    assert X.shape == Y.shape, "Matrices are not aligned."
    eps = np.finfo(Y.dtype).eps if not 'int' in str(Y.dtype) else 0
    return np.mat([[op(X[i, j], Y[i, j] + eps) for j in range(X.shape[1])] for i in range(X.shape[0])])


def inf_norm(X):
    """
    Infinity norm of a matrix (maximum absolute row sum).

    :param X: Input matrix.
    :type X: :class:`scipy.sparse.csr_matrix`, :class:`scipy.sparse.csc_matrix`
    or :class:`numpy.matrix`
    """
    if sp.isspmatrix_csr(X) or sp.isspmatrix_csc(X):
        # avoid copying index and ptr arrays
        abs_X = X.__class__(
            (abs(X.data), X.indices, X.indptr), shape=X.shape)
        return (abs_X * np.ones((X.shape[1]), dtype=X.dtype)).max()
    elif sp.isspmatrix(X):
        return (abs(X) * np.ones((X.shape[1]), dtype=X.dtype)).max()
    else:
        return nla.norm(np.asmatrix(X), float('inf'))


def norm(X, p="fro"):
    """
    Compute entry-wise norms (! not induced/operator norms).

    :param X: The input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param p: Order of the norm.
    :type p: `str` or `float`
    """
    assert 1 in X.shape or p != 2, "Computing entry-wise norms only."
    if sp.isspmatrix(X):
        fro = lambda X: sum(abs(x) ** 2 for x in X.data) ** (1. / 2)
        inf = lambda X: abs(X).sum(
            axis=1).max() if 1 not in X.shape else abs(X).max()
        m_inf = lambda X: abs(X).sum(
            axis=1).min() if 1 not in X.shape else abs(X).min()
        one = lambda X: abs(X).sum(axis=0).max() if 1 not in X.shape else sum(
            abs(x) ** p for x in X.data) ** (1. / p)
        m_one = lambda X: abs(X).sum(axis=0).min() if 1 not in X.shape else sum(
            abs(x) ** p for x in X.data) ** (1. / p)
        v = {
            "fro": fro,
            "inf": inf,
            "-inf": m_inf,
            1: one,
            -1: m_one,
        }.get(p)
        return v(X) if v != None else sum(abs(x) ** p for x in X.data) ** (1. / p)
    else:
        return nla.norm(np.mat(X), p)


def vstack(X, format=None, dtype=None):
    """
    Stack sparse or dense matrices vertically (row wise).

    :param X: Sequence of matrices with compatible shapes.
    :type X: sequence of :class:`scipy.sparse` of format csr, csc, coo, bsr,
    dok, lil, dia or :class:`numpy.matrix`
    """
    if len([0 for x in X if not sp.isspmatrix(x)]) == 0:
        # scipy.sparse bug
        # return sp.vstack(X, format = X[0].getformat() if format is None else
        # format, dtype = X[0].dtype if dtype is None else dtype)
        return sp.vstack(X)
    else:
        return np.vstack(X)


def hstack(X, format=None, dtype=None):
    """
    Stack sparse or dense matrices horizontally (column wise).

    :param X: Sequence of matrices with compatible shapes.
    :type X: sequence of :class:`scipy.sparse` of format csr, csc, coo, bsr,
    dok, lil, dia or :class:`numpy.matrix`
    """
    if len([0 for x in X if not sp.isspmatrix(x)]) == 0:
        # scipy.sparse bug
        # return sp.hstack(X, format = X[0].getformat() if format is None else
        # format, dtype = X[0].dtyoe if dtype is None else dtype)
        return sp.hstack(X)
    else:
        return np.hstack(X)


def max(X, s):
    """
    Compute element-wise max(x,s) assignment for sparse or dense matrix.

    :param X: The input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param s: the input scalar.
    :type s: `float`
    """
    if sp.isspmatrix(X):
        Y = X.tocsr()
        DD = Y.data.copy()
        DD = np.maximum(DD, s)
        return sp.csr_matrix((DD, Y.indices, Y.indptr), Y.shape)
    else:
        return np.maximum(X, s)


def min(X, s):
    """
    Compute element-wise min(x,s) assignment for sparse or dense matrix.

    :param X: The input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param s: the input scalar.
    :type s: `float`
    """
    if sp.isspmatrix(X):
        Y = X.tocsr()
        DD = Y.data.copy()
        DD = np.minimum(DD, s)
        return sp.csr_matrix((DD, Y.indices, Y.indptr), Y.shape)
    else:
        return np.minimum(X, s)


def count(X, s):
    """
    Return the number of occurrences of element :param:`s` in sparse or
    dense matrix X.

    :param X: The input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param s: the input scalar.
    :type s: `float`
    """
    if sp.isspmatrix(X):
        return sum([1 for x in X.data if s == x])
    else:
        return sum([1 for r in X.tolist() for x in r if s == x])


def nz_data(X):
    """
    Return list of nonzero elements from X (! data, not indices).

    :param X: The input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    """
    if sp.isspmatrix(X):
        return X.data.tolist()
    else:
        return [x for r in X.tolist() for x in r if x != 0]


def choose(n, k):
    """
    A fast way to calculate binomial coefficients C(n, k). It is 10 times faster
    than scipy.mis.comb for exact answers.

    :param n: Index of binomial coefficient.
    :type n: `int`
    :param k: Index of binomial coefficient.
    :type k: `int`
    """
    if 0 <= k <= n:
        ntok = 1
        ktok = 1
        for t in range(1, min(k, n - k) + 1):
            ntok *= n
            ktok *= t
            n -= 1
        return ntok // ktok
    else:
        return 0




"""
###################################
Pmf (``methods.factorization.pmf``)
###################################

**Probabilistic Nonnegative Matrix Factorization (PMF).** 

PMF interprets target matrix (V) as samples from a multinomial [Laurberg2008]_,
[Hansen2008]_ and uses Euclidean distance for convergence test. Factorization
is guided by an expectation maximization algorithm.

.. literalinclude:: /code/snippet_pmf.py

"""

from nimfa.models import *
from nimfa.utils import *
from nimfa.utils.linalg import *

__all__ = ['Pmf']


class Pmf(nmf_std.Nmf_std):
    """
    :param V: The target matrix to estimate.
    :type V: Instance of the :class:`scipy.sparse` sparse matrices types,
       :class:`numpy.ndarray`, :class:`numpy.matrix` or tuple of instances of
       the latter classes.

    :param seed: Specify method to seed the computation of a factorization. If
       specified :param:`W` and :param:`H` seeding must be None. If neither seeding
       method or initial fixed factorization is specified, random initialization is
       used.
    :type seed: `str` naming the method or :class:`methods.seeding.nndsvd.Nndsvd`
       or None

    :param W: Specify initial factorization of basis matrix W. Default is None.
       When specified, :param:`seed` must be None.
    :type W: :class:`scipy.sparse` or :class:`numpy.ndarray` or
       :class:`numpy.matrix` or None

    :param H: Specify initial factorization of mixture matrix H. Default is None.
       When specified, :param:`seed` must be None.
    :type H: Instance of the :class:`scipy.sparse` sparse matrices types,
       :class:`numpy.ndarray`, :class:`numpy.matrix`, tuple of instances of the
       latter classes or None

    :param rank: The factorization rank to achieve. Default is 30.
    :type rank: `int`

    :param n_run: It specifies the number of runs of the algorithm. Default is
       1. If multiple runs are performed, fitted factorization model with the
       lowest objective function value is retained.
    :type n_run: `int`

    :param callback: Pass a callback function that is called after each run when
       performing multiple runs. This is useful if one wants to save summary
       measures or process the result before it gets discarded. The callback
       function is called with only one argument :class:`models.mf_fit.Mf_fit` that
       contains the fitted model. Default is None.
    :type callback: `function`

    :param callback_init: Pass a callback function that is called after each
       initialization of the matrix factors. In case of multiple runs the function
       is called before each run (more precisely after initialization and before
       the factorization of each run). In case of single run, the passed callback
       function is called after the only initialization of the matrix factors.
       This is useful if one wants to obtain the initialized matrix factors for
       further analysis or additional info about initialized factorization model.
       The callback function is called with only one argument
       :class:`models.mf_fit.Mf_fit` that (among others) contains also initialized
       matrix factors. Default is None.
    :type callback_init: `function`

    :param track_factor: When :param:`track_factor` is specified, the fitted
        factorization model is tracked during multiple runs of the algorithm. This
        option is taken into account only when multiple runs are executed
        (:param:`n_run` > 1). From each run of the factorization all matrix factors
        are retained, which can be very space consuming. If space is the problem
        setting the callback function with :param:`callback` is advised which is
        executed after each run. Tracking is useful for performing some quality or
        performance measures (e.g. cophenetic correlation, consensus matrix,
        dispersion). By default fitted model is not tracked.
    :type track_factor: `bool`

    :param track_error: Tracking the residuals error. Only the residuals from
        each iteration of the factorization are retained. Error tracking is not
        space consuming. By default residuals are not tracked and only the final
        residuals are saved. It can be used for plotting the trajectory of the
        residuals.
    :type track_error: `bool`

    :param rel_error: In PMF only Euclidean distance cost function is used for
       convergence test by default. By specifying the value for minimum relative
       error, the relative error measure can be used as stopping criteria as well.
       In this case of multiple passed criteria, the satisfiability of one terminates
       the factorization run. Suggested value for ``rel_error`` is 1e-5.
    :type rel_error: `float`

    **Stopping criterion**

    Factorization terminates if any of specified criteria is satisfied.

    :param max_iter: Maximum number of factorization iterations. Note that the
       number of iterations depends on the speed of method convergence. Default
       is 30.
    :type max_iter: `int`

    :param min_residuals: Minimal required improvement of the residuals from the
       previous iteration. They are computed between the target matrix and its MF
       estimate using the objective function associated to the MF algorithm.
       Default is None.
    :type min_residuals: `float`

    :param test_conv: It indicates how often convergence test is done. By
       default convergence is tested each iteration.
    :type test_conv: `int`
    """
    def __init__(self, V, seed=None, W=None, H=None, rank=30, max_iter=30,
                 min_residuals=1e-5, test_conv=None, n_run=1, callback=None,
                 callback_init=None, track_factor=False, track_error=False,
                 rel_error=False, **options):
        self.name = "pmf"
        self.aseeds = ["random", "fixed", "nndsvd", "random_c", "random_vcol"]
        nmf_std.Nmf_std.__init__(self, vars())
        self.tracker = mf_track.Mf_track() if self.track_factor and self.n_run > 1 \
                                              or self.track_error else None

    def factorize(self):
        """
        Compute matrix factorization.
         
        Return fitted factorization model.
        """
        for run in range(self.n_run):
            self.W, self.H = self.seed.initialize(
                self.V, self.rank, self.options)
            self.W = elop(
                self.W, repmat(self.W.sum(axis=0), self.V.shape[0], 1), div)
            self.H = elop(
                self.H, repmat(self.H.sum(axis=1), 1, self.V.shape[1]), div)
            self.v_factor = self.V.sum()
            self.V_n = sop(self.V.copy(), self.v_factor, div)
            self.P = np.diag([1. / self.rank for _ in range(self.rank)])
            self.sqrt_P = sop(self.P, s=None, op=np.sqrt)
            p_obj = c_obj = sys.float_info.max
            best_obj = c_obj if run == 0 else best_obj
            self.error_v_n = c_obj
            iter = 0
            if self.callback_init:
                self.final_obj = c_obj
                self.n_iter = iter
                mffit = mf_fit.Mf_fit(self)
                self.callback_init(mffit)
            while self.is_satisfied(p_obj, c_obj, iter):
                p_obj = c_obj if not self.test_conv or iter % self.test_conv == 0 else p_obj
                self.update()
                self._adjustment()
                iter += 1
                c_obj = self.objective(
                ) if not self.test_conv or iter % self.test_conv == 0 else c_obj
                if self.track_error:
                    self.tracker.track_error(run, c_obj)
            self.W = self.v_factor * dot(self.W, self.sqrt_P)
            self.H = dot(self.sqrt_P, self.H)
            if self.callback:
                self.final_obj = c_obj
                self.n_iter = iter
                mffit = mf_fit.Mf_fit(self)
                self.callback(mffit)
            if self.track_factor:
                self.tracker.track_factor(
                    run, W=self.W, H=self.H, final_obj=c_obj, n_iter=iter)
            # if multiple runs are performed, fitted factorization model with
            # the lowest objective function value is retained
            if c_obj <= best_obj or run == 0:
                best_obj = c_obj
                self.n_iter = iter
                self.final_obj = c_obj
                mffit = mf_fit.Mf_fit(copy.deepcopy(self))

        mffit.fit.tracker = self.tracker
        return mffit

    def is_satisfied(self, p_obj, c_obj, iter):
        """
        Compute the satisfiability of the stopping criteria based on stopping
        parameters and objective function value.
        
        Return logical value denoting factorization continuation. 
        
        :param p_obj: Objective function value from previous iteration. 
        :type p_obj: `float`
        :param c_obj: Current objective function value.
        :type c_obj: `float`
        :param iter: Current iteration number. 
        :type iter: `int`
        """
        if self.max_iter and self.max_iter <= iter:
            return False
        if self.test_conv and iter % self.test_conv != 0:
            return True
        if self.min_residuals and iter > 0 and p_obj - c_obj < self.min_residuals:
            return False
        if iter > 0 and c_obj > p_obj:
            return False
        if self.rel_error and self.error_v_n < self.rel_error:
            return False
        return True

    def _adjustment(self):
        """Adjust small values to factors to avoid numerical underflow."""
        self.H = max(self.H, np.finfo(self.H.dtype).eps)
        self.W = max(self.W, np.finfo(self.W.dtype).eps)

    def update(self):
        """Update basis and mixture matrix. It is expectation maximization
        algorithm. """
        # E step
        Qnorm = dot(dot(self.W, self.P), self.H)
        for k in range(self.rank):
            # E-step
            Q = elop(self.P[k, k] * dot(self.W[:, k], self.H[k, :]), sop(
                Qnorm, np.finfo(Qnorm.dtype).eps, add), div)
            V_nQ = multiply(self.V_n, Q)
            # M-step
            dum = V_nQ.sum(axis=1)
            s_dum = dum.sum()
            for i in range(self.W.shape[0]):
                self.W[i, k] = dum[i, 0] / s_dum
            dum = V_nQ.sum(axis=0)
            s_dum = dum.sum()
            for i in range(self.H.shape[1]):
                self.H[k, i] = dum[0, i] / s_dum

    def objective(self):
        """Compute Euclidean distance cost function."""
        # relative error
        self.error_v_n = abs(
            self.V_n - dot(self.W, self.H)).mean() / self.V_n.mean()
        # Euclidean distance
        return power(self.V - dot(dot(dot(self.W, self.sqrt_P) * self.v_factor, self.sqrt_P), self.H), 2).sum()

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name




"""    
This package contains factorization models used in the library nimfa. Specifically, it contains the following:

    #. Generic factorization model for handling common computations and assessing quality and performance
       measures in NMF.
    #. Generic factorization model for handling standard MF.
    #. Implementation of the standard model to manage factorizations that follow standard NMF model.
    #. Implementation of the alternative nonsmooth model to manage factorizations that follow nonstandard NMF model
       (e.g. nsnmf).
    #. Implementation of the alternative multiple model to manage factorizations that follow NMF nonstandard model
       (e.g. snmnmf).
    #. Tracker of MF fitted model across multiple runs of the factorizations and tracker of
       the residuals error across iterations of single/multiple runs.
    #. Implementation for storing MF fitted model and MF results.
"""

from . import nmf
from . import nmf_std
from . import nmf_ns
from . import nmf_mm
from . import smf
from . import mf_track
from . import mf_fit




"""
#####################################
Snmf (``methods.factorization.snmf``)
#####################################

**Sparse Nonnegative Matrix Factorization (SNMF)** based on alternating
nonnegativity constrained least squares [Park2007]_.

In order to enforce sparseness on basis or mixture matrix, SNMF can be used,
namely two formulations: SNMF/L for sparse W (sparseness is imposed on the left
factor) and SNMF/R for sparse H (sparseness imposed on the right factor). These
formulations utilize L1-norm minimization. Each subproblem is solved by a fast
nonnegativity constrained least squares (FCNNLS) algorithm (van Benthem and
Keenan, 2004) that is improved upon the active set based NLS method.

SNMF/R contains two subproblems for two-block minimization scheme. The objective
function is coercive on the feasible set. It can be shown (Grippo and Sciandrome,
2000) that two-block minimization process is convergent, every accumulation point
is a critical point of the corresponding problem. Similarly, the algorithm SNMF/L
converges to a stationary point.

.. literalinclude:: /code/snippet_snmf_l.py

.. literalinclude:: /code/snippet_snmf_r.py

"""

from nimfa.models import *
from nimfa.utils import *
from nimfa.utils.linalg import *

__all__ = ['Snmf']


class Snmf(nmf_std.Nmf_std):
    """
    :param V: The target matrix to estimate.
    :type V: Instance of the :class:`scipy.sparse` sparse matrices types,
       :class:`numpy.ndarray`, :class:`numpy.matrix` or tuple of instances of
       the latter classes.

    :param seed: Specify method to seed the computation of a factorization. If
       specified :param:`W` and :param:`H` seeding must be None. If neither seeding
       method or initial fixed factorization is specified, random initialization is
       used.
    :type seed: `str` naming the method or :class:`methods.seeding.nndsvd.Nndsvd`
       or None

    :param W: Specify initial factorization of basis matrix W. Default is None.
       When specified, :param:`seed` must be None.
    :type W: :class:`scipy.sparse` or :class:`numpy.ndarray` or
       :class:`numpy.matrix` or None

    :param H: Specify initial factorization of mixture matrix H. Default is None.
       When specified, :param:`seed` must be None.
    :type H: Instance of the :class:`scipy.sparse` sparse matrices types,
       :class:`numpy.ndarray`, :class:`numpy.matrix`, tuple of instances of the
       latter classes or None

    :param rank: The factorization rank to achieve. Default is 30.
    :type rank: `int`

    :param n_run: It specifies the number of runs of the algorithm. Default is
       1. If multiple runs are performed, fitted factorization model with the
       lowest objective function value is retained.
    :type n_run: `int`

    :param callback: Pass a callback function that is called after each run when
       performing multiple runs. This is useful if one wants to save summary
       measures or process the result before it gets discarded. The callback
       function is called with only one argument :class:`models.mf_fit.Mf_fit` that
       contains the fitted model. Default is None.
    :type callback: `function`

    :param callback_init: Pass a callback function that is called after each
       initialization of the matrix factors. In case of multiple runs the function
       is called before each run (more precisely after initialization and before
       the factorization of each run). In case of single run, the passed callback
       function is called after the only initialization of the matrix factors.
       This is useful if one wants to obtain the initialized matrix factors for
       further analysis or additional info about initialized factorization model.
       The callback function is called with only one argument
       :class:`models.mf_fit.Mf_fit` that (among others) contains also initialized
       matrix factors. Default is None.
    :type callback_init: `function`

    :param track_factor: When :param:`track_factor` is specified, the fitted
        factorization model is tracked during multiple runs of the algorithm. This
        option is taken into account only when multiple runs are executed
        (:param:`n_run` > 1). From each run of the factorization all matrix factors
        are retained, which can be very space consuming. If space is the problem
        setting the callback function with :param:`callback` is advised which is
        executed after each run. Tracking is useful for performing some quality or
        performance measures (e.g. cophenetic correlation, consensus matrix,
        dispersion). By default fitted model is not tracked.
    :type track_factor: `bool`

    :param track_error: Tracking the residuals error. Only the residuals from
        each iteration of the factorization are retained. Error tracking is not
        space consuming. By default residuals are not tracked and only the final
        residuals are saved. It can be used for plotting the trajectory of the
        residuals.
    :type track_error: `bool`

    :param version: Specifiy version of the SNMF algorithm. it has two accepting
       values, 'r' and 'l' for SNMF/R and SNMF/L, respectively. Default choice is
       SNMF/R.
    :type version: `str`

    :param eta: Used for suppressing Frobenius norm on the basis matrix (W).
       Default value is maximum value of the target matrix (V). If ``eta`` is
       negative, maximum value of target matrix is used for it.
    :type eta: `float`

    :param beta: It controls sparseness. Larger ``beta`` generates higher
       sparseness on H. Too large :param:`beta` is not recommended. It should have
       positive value. Default value is 1e-4.
    :type beta: `float`

    :param i_conv: Part of the biclustering convergence test. It decides
       convergence if row clusters and column clusters have not changed for
       ``i_conv`` convergence tests. It should have nonnegative value. Default
       value is 10.
    :type i_conv: `int`

    :param w_min_change: Part of the biclustering convergence test. It specifies
       the minimal allowance of the change of row clusters. It should have
       nonnegative value. Default value is 0.
    :type w_min_change: `int`

    **Stopping criterion**

    Factorization terminates if any of specified criteria is satisfied.

    :param max_iter: Maximum number of factorization iterations. Note that the
       number of iterations depends on the speed of method convergence. Default
       is 30.
    :type max_iter: `int`

    :param min_residuals: Minimal required improvement of the residuals from the
       previous iteration. They are computed between the target matrix and its MF
       estimate using the objective function associated to the MF algorithm.
       Default is None.
    :type min_residuals: `float`

    :param test_conv: It indicates how often convergence test is done. By
       default convergence is tested each iteration.
    :type test_conv: `int`
    """
    def __init__(self, V, seed=None, W=None, H=None, H1=None,
                 rank=30, max_iter=30, min_residuals=1e-5, test_conv=None,
                 n_run=1, callback=None, callback_init=None, track_factor=False,
                 track_error=False, version='r', eta=None, beta=1e-4,
                 i_conv=10, w_min_change=0, **options):
        self.name = "snmf"
        self.aseeds = ["random", "fixed", "nndsvd", "random_c", "random_vcol"]
        nmf_std.Nmf_std.__init__(self, vars())
        if not self.eta:
            self.eta = np.max(self.V) if not sp.isspmatrix(self.V) else np.max(self.V.data)
        if self.eta < 0:
            self.eta = np.max(self.V) if not sp.isspmatrix(self.V) else 0.
        self.min_residuals = 1e-4 if not self.min_residuals else self.min_residuals
        self.tracker = mf_track.Mf_track() if self.track_factor and self.n_run > 1 \
                                              or self.track_error else None

    def factorize(self):
        """
        Compute matrix factorization. 
                
        Return fitted factorization model.
        """
        for run in range(self.n_run):
            self.W, self.H = self.seed.initialize(self.V, self.rank, self.options)
            # in version SNMF/L, V is transposed while W and H are swapped and
            # transposed.
            if self.version == 'l':
                self.V = self.V.T
                self.W, self.H = self.H.T, self.W.T

            if sp.isspmatrix(self.W):
                self.W = self.W.tolil()
            if sp.isspmatrix(self.H):
                self.H = self.H.tolil()
            iter = 0
            self.idx_w_old = np.mat(np.zeros((self.V.shape[0], 1)))
            self.idx_h_old = np.mat(np.zeros((1, self.V.shape[1])))
            c_obj = sys.float_info.max
            best_obj = c_obj if run == 0 else best_obj
            # count the number of convergence checks that column clusters and
            # row clusters have not changed.
            self.inc = 0
            # normalize W
            w_c = sop(multiply(self.W, self.W).sum(axis=0), op=np.sqrt)
            self.W = elop(self.W, repmat(w_c, self.V.shape[0], 1), div)
            if sp.isspmatrix(self.V):
                self.beta_vec = sqrt(self.beta) * sp.lil_matrix(
                    np.ones((1, self.rank)), dtype=self.V.dtype)
                self.I_k = self.eta * \
                    sp.eye(self.rank, self.rank, format='lil')
            else:
                self.beta_vec = sqrt(self.beta) * np.ones((1, self.rank))
                self.I_k = self.eta * np.mat(np.eye(self.rank))
            self.n_restart = 0
            if self.callback_init:
                self.final_obj = c_obj
                self.n_iter = iter
                mffit = mf_fit.Mf_fit(self)
                self.callback_init(mffit)
            while self.is_satisfied(c_obj, iter):
                self.update()
                iter += 1
                c_obj = self.objective(
                ) if not self.test_conv or iter % self.test_conv == 0 else c_obj
                if self.track_error:
                    self.tracker.track_error(run, c_obj)
            # basis and mixture matrix are now constructed and are now
            # converted to CSR for fast LA operations
            if sp.isspmatrix(self.W):
                self.W = self.W.tocsr()
            if sp.isspmatrix(self.H):
                self.H = self.H.tocsr()
            # transpose and swap factors if SNMF/L
            if self.version == 'l':
                self.V = self.V.T
                self.W, self.H = self.H.T, self.W.T
            if self.callback:
                self.final_obj = c_obj
                self.n_iter = iter
                mffit = mf_fit.Mf_fit(self)
                self.callback(mffit)
            if self.track_factor:
                self.tracker.track_factor(
                    run, W=self.W, H=self.H, final_obj=c_obj, n_iter=iter)
            # if multiple runs are performed, fitted factorization model with
            # the lowest objective function value is retained
            if c_obj <= best_obj or run == 0:
                best_obj = c_obj
                self.n_iter = iter
                self.final_obj = c_obj
                mffit = mf_fit.Mf_fit(copy.deepcopy(self))

        mffit.fit.tracker = self.tracker
        return mffit

    def is_satisfied(self, c_obj, iter):
        """
        Compute the satisfiability of the stopping criteria based on stopping
        parameters and objective function value.
        
        Return logical value denoting factorization continuation. 
        
        :param c_obj: Current objective function value.
        :type c_obj: `float`
        :param iter: Current iteration number. 
        :type iter: `int`
        """
        if iter == 0:
            self.init_erravg = c_obj
        if self.max_iter and self.max_iter <= iter:
            return False
        if self.test_conv and iter % self.test_conv != 0:
            return True
        if self.inc >= self.i_conv and c_obj < self.min_residuals * self.init_erravg:
            return False
        return True

    def update(self):
        """Update basis and mixture matrix."""
        if sp.isspmatrix(self.V):
            v1 = self.V.__class__((1, self.V.shape[1]), dtype=self.V.dtype)
            v1t = self.V.__class__(
                (self.rank, self.V.shape[0]), dtype=self.V.dtype)
        else:
            v1 = np.zeros((1, self.V.shape[1]))
            v1t = np.zeros((self.rank, self.V.shape[0]))
        # min_h ||[[W; 1 ... 1]*H  - [A; 0 ... 0]||, s.t. H>=0, for given A and
        # W
        if sp.isspmatrix(self.V):
            self.H = self._spfcnnls(
                vstack((self.W, self.beta_vec)), vstack((self.V, v1)))
        else:
            self.H = self._fcnnls(
                vstack((self.W, self.beta_vec)), vstack((self.V, v1)))
        if any(self.H.sum(axis=1) == 0):
            self.n_restart += 1
            if self.n_restart >= 100:
                raise utils.MFError(
                    "Too many restarts due to too large beta parameter.")
            self.idx_w_old = np.mat(np.zeros((self.V.shape[0], 1)))
            self.idx_h_old = np.mat(np.zeros((1, self.V.shape[1])))
            self.inc = 0
            self.W, _ = self.seed.initialize(self.V, self.rank, self.options)
            # normalize W and convert to lil
            w_c = sop(multiply(self.W, self.W).sum(axis=0), op=np.sqrt)
            if sp.issparse(self.W):
                self.W = elop(self.W, repmat(w_c, self.V.shape[0], 1), div).tolil()
            else:
                self.W = elop(self.W, repmat(w_c, self.V.shape[0], 1), div)
            return
        # min_w ||[H'; I_k]*W' - [A'; 0]||, s.t. W>=0, for given A and H.
        if sp.isspmatrix(self.V):
            Wt = self._spfcnnls(
                vstack((self.H.T, self.I_k)), vstack((self.V.T, v1t)))
        else:
            Wt = self._fcnnls(
                vstack((self.H.T, self.I_k)), vstack((self.V.T, v1t)))
        self.W = Wt.T

    def fro_error(self):
        """Compute NMF objective value with additional sparsity constraints."""
        rec_err = norm(self.V - dot(self.W, self.H), "fro")
        w_norm = norm(self.W, "fro")
        hc_norm = sum(norm(self.H[:, j], 1) ** 2 for j in self.H.shape[1])
        return 0.5 * rec_err ** 2 + self.eta * w_norm ** 2 + self.beta * hc_norm

    def objective(self):
        """Compute convergence test."""
        _, idx_w = argmax(self.W, axis=1)
        _, idx_h = argmax(self.H, axis=0)
        changed_w = count(elop(idx_w, self.idx_w_old, ne), 1)
        changed_h = count(elop(idx_h, self.idx_h_old, ne), 1)
        if changed_w <= self.w_min_change and changed_h == 0:
            self.inc += 1
        else:
            self.inc = 0
        WtWH = dot(dot(self.W.T, self.W), self.H)
        WtV = dot(self.W.T, self.V)
        bRH = dot(self.beta * np.ones((self.rank, self.rank)), self.H)
        resmat = elop(self.H, WtWH - WtV + bRH, min)
        WHHt = dot(self.W, dot(self.H, self.H.T))
        VHt = dot(self.V, self.H.T)
        resmat1 = elop(self.W, WHHt - VHt + self.eta ** 2 * self.W, min)
        res_vec = nz_data(resmat) + nz_data(resmat1)
        # L1 norm
        self.conv = norm(np.mat(res_vec), 1)
        err_avg = self.conv / len(res_vec)
        self.idx_w_old = idx_w
        self.idx_h_old = idx_h
        return err_avg

    def _spfcnnls(self, C, A):
        """
        NNLS for sparse matrices.
        
        Nonnegative least squares solver (NNLS) using normal equations and fast
        combinatorial strategy (van Benthem and Keenan, 2004).
        
        Given A and C this algorithm solves for the optimal K in a least squares
        sense, using that A = C*K in the problem ||A - C*K||, s.t.
        K>=0 for given A and C.
        
        C is the n_obs x l_var coefficient matrix
        A is the n_obs x p_rhs matrix of observations
        K is the l_var x p_rhs solution matrix
        
        p_set is set of passive sets, one for each column. 
        f_set is set of column indices for solutions that have not yet converged. 
        h_set is set of column indices for currently infeasible solutions. 
        j_set is working set of column indices for currently optimal solutions. 
        """
        C = C.tolil()
        A = A.tolil()
        _, l_var = C.shape
        p_rhs = A.shape[1]
        W = sp.lil_matrix((l_var, p_rhs))
        iter = 0
        max_iter = 3 * l_var
        # precompute parts of pseudoinverse
        CtC = dot(C.T, C)
        CtA = dot(C.T, A)
        # obtain the initial feasible solution and corresponding passive set
        K = self.__spcssls(CtC, CtA)
        p_set = sop(K, 0, ge).tolil()
        for i in range(K.shape[0]):
            for j in range(K.shape[1]):
                if not p_set[i, j]:
                    K[i, j] = 0.
        D = K.copy()
        f_set = np.array(find(np.logical_not(all(p_set, axis=0))))
        # active set algorithm for NNLS main loop
        while len(f_set) > 0:
            # solve for the passive variables
            K[:, f_set] = self.__spcssls(CtC, CtA[:, f_set], p_set[:, f_set])
            # find any infeasible solutions
            idx = find(any(sop(K[:, f_set], 0, le), axis=0))
            h_set = f_set[idx] if idx != [] else []
            # make infeasible solutions feasible (standard NNLS inner loop)
            if len(h_set) > 0:
                n_h_set = len(h_set)
                alpha = np.mat(np.zeros((l_var, n_h_set)))
                while len(h_set) > 0 and iter < max_iter:
                    iter += 1
                    alpha[:, :n_h_set] = np.Inf
                    # find indices of negative variables in passive set
                    tmp = sop(K[:, h_set], 0, le).tolil()
                    tmp_f = sp.lil_matrix(K.shape, dtype='bool')
                    for i in range(K.shape[0]):
                        for j in range(len(h_set)):
                            if p_set[i, h_set[j]] and tmp[i, h_set[j]]:
                                tmp_f[i, h_set[j]] = True
                    idx_f = find(tmp_f[:, h_set])
                    i_f = [l % p_set.shape[0] for l in idx_f]
                    j_f = [l // p_set.shape[0] for l in idx_f]
                    if len(i_f) == 0:
                        break
                    if n_h_set == 1:
                        h_n = h_set * np.ones((1, len(j_f)))
                        l_1n = i_f
                        l_2n = h_n.tolist()[0]
                    else:
                        l_1n = i_f
                        l_2n = [h_set[e] for e in j_f]
                    t_d = D[l_1n, l_2n] / (D[l_1n, l_2n] - K[l_1n, l_2n])
                    for i in range(len(i_f)):
                        alpha[i_f[i], j_f[i]] = t_d.todense().flatten()[0, i]
                    alpha_min, min_idx = argmin(alpha[:, :n_h_set], axis=0)
                    min_idx = min_idx.tolist()[0]
                    alpha[:, :n_h_set] = repmat(alpha_min, l_var, 1)
                    D[:, h_set] = D[:, h_set] - multiply(
                        alpha[:, :n_h_set], D[:, h_set] - K[:, h_set])
                    D[min_idx, h_set] = 0
                    p_set[min_idx, h_set] = 0
                    K[:, h_set] = self.__spcssls(
                        CtC, CtA[:, h_set], p_set[:, h_set])
                    h_set = find(any(sop(K, 0, le), axis=0))
                    n_h_set = len(h_set)
            # make sure the solution has converged and check solution for
            # optimality
            W[:, f_set] = CtA[:, f_set] - dot(CtC, K[:, f_set])
            tmp = sp.lil_matrix(p_set.shape, dtype='bool')
            for i in range(p_set.shape[0]):
                for j in f_set:
                    if not p_set[i, j]:
                        tmp[i, j] = True
            p_set_not = np.logical_not(p_set[:, f_set].todense())
            w = W[:, f_set].todense()
            j_set = find(all(multiply(p_set_not, w) <= 0, axis=0))
            f_j = f_set[j_set] if j_set != [] else []
            f_set = np.setdiff1d(np.asarray(f_set), np.asarray(f_j))
            # for non-optimal solutions, add the appropriate variable to Pset
            if len(f_set) > 0:
                tmp = sp.lil_matrix(p_set.shape, dtype='bool')
                for i in range(p_set.shape[0]):
                    for j in f_set:
                        if not p_set[i, j]:
                            tmp[i, j] = True
                _, mxidx = argmax(multiply(tmp[:, f_set], W[:, f_set]), axis=0)
                mxidx = mxidx.tolist()[0]
                p_set[mxidx, f_set] = 1
                D[:, f_set] = K[:, f_set]
        return K.tolil()

    def __spcssls(self, CtC, CtA, p_set=None):
        """
        Solver for sparse matrices.
        
        Solve the set of equations CtA = CtC * K for variables defined in set p_set
        using the fast combinatorial approach (van Benthem and Keenan, 2004).
        
        It returns matrix in LIL sparse format.
        """
        K = sp.lil_matrix(CtA.shape)
        if p_set == None or p_set.size == 0 or all(p_set):
            # equivalent if CtC is square matrix
            for k in range(CtA.shape[1]):
                ls = sp.linalg.gmres(CtC, CtA[:, k].toarray())[0]
                K[:, k] = sp.lil_matrix(np.mat(ls).T)
            # K = dot(np.linalg.pinv(CtC), CtA)
        else:
            l_var, p_rhs = p_set.shape
            coded_p_set = dot(
                sp.lil_matrix(np.mat(2 ** np.array(list(range(l_var - 1, -1, -1))))), p_set)
            sorted_p_set, sorted_idx_set = sort(coded_p_set.todense())
            breaks = diff(np.mat(sorted_p_set))
            break_idx = [-1] + find(np.mat(breaks)) + [p_rhs]
            for k in range(len(break_idx) - 1):
                cols2solve = sorted_idx_set[
                    break_idx[k] + 1: break_idx[k + 1] + 1]
                vars = p_set[:, sorted_idx_set[break_idx[k] + 1]]
                vars = [i for i in range(vars.shape[0]) if vars[i, 0]]
                tmp_ls = CtA[:, cols2solve][vars, :]
                sol = sp.lil_matrix(K.shape)
                for k in range(tmp_ls.shape[1]):
                    ls = sp.linalg.gmres(CtC[:, vars][vars, :], tmp_ls[:, k].toarray())[0]
                    sol[:, k] = sp.lil_matrix(np.mat(ls).T)
                i = 0
                for c in cols2solve:
                    j = 0
                    for v in vars:
                        K[v, c] = sol[j, i]
                        j += 1
                    i += 1
                # K[vars, cols2solve] = dot(np.linalg.pinv(CtC[vars, vars]),
                # CtA[vars, cols2solve])
        return K.tolil()

    def _fcnnls(self, C, A):
        """
        NNLS for dense matrices.
        
        Nonnegative least squares solver (NNLS) using normal equations and fast
        combinatorial strategy (van Benthem and Keenan, 2004).
        
        Given A and C this algorithm solves for the optimal K in a least squares
        sense, using that A = C*K in the problem
        ||A - C*K||, s.t. K>=0 for given A and C. 
        
        C is the n_obs x l_var coefficient matrix
        A is the n_obs x p_rhs matrix of observations
        K is the l_var x p_rhs solution matrix
        
        p_set is set of passive sets, one for each column. 
        f_set is set of column indices for solutions that have not yet converged. 
        h_set is set of column indices for currently infeasible solutions. 
        j_set is working set of column indices for currently optimal solutions. 
        """
        C = C.todense() if sp.isspmatrix(C) else C
        A = A.todense() if sp.isspmatrix(A) else A
        _, l_var = C.shape
        p_rhs = A.shape[1]
        W = np.mat(np.zeros((l_var, p_rhs)))
        iter = 0
        max_iter = 3 * l_var
        # precompute parts of pseudoinverse
        CtC = dot(C.T, C)
        CtA = dot(C.T, A)
        # obtain the initial feasible solution and corresponding passive set
        # K is not sparse
        K = self.__cssls(CtC, CtA)
        p_set = K > 0
        K[np.logical_not(p_set)] = 0
        D = K.copy()
        f_set = np.array(find(np.logical_not(all(p_set, axis=0))))
        # active set algorithm for NNLS main loop
        while len(f_set) > 0:
            # solve for the passive variables
            K[:, f_set] = self.__cssls(CtC, CtA[:, f_set], p_set[:, f_set])
            # find any infeasible solutions
            idx = find(any(K[:, f_set] < 0, axis=0))
            h_set = f_set[idx] if idx != [] else []
            # make infeasible solutions feasible (standard NNLS inner loop)
            if len(h_set) > 0:
                n_h_set = len(h_set)
                alpha = np.mat(np.zeros((l_var, n_h_set)))
                while len(h_set) > 0 and iter < max_iter:
                    iter += 1
                    alpha[:, :n_h_set] = np.Inf
                    # find indices of negative variables in passive set
                    idx_f = find(
                        np.logical_and(p_set[:, h_set], K[:, h_set] < 0))
                    i_f = [l % p_set.shape[0] for l in idx_f]
                    j_f = [l // p_set.shape[0] for l in idx_f]
                    if len(i_f) == 0:
                        break
                    if n_h_set == 1:
                        h_n = h_set * np.ones((1, len(j_f)))
                        l_1n = i_f
                        l_2n = list(map(int, h_n.tolist()[0]))
                    else:
                        l_1n = i_f
                        l_2n = list(map(int, [h_set[e] for e in j_f]))
                    t_d = D[l_1n, l_2n] / (D[l_1n, l_2n] - K[l_1n, l_2n])
                    for i in range(len(i_f)):
                        alpha[i_f[i], j_f[i]] = t_d.flatten()[0, i]
                    alpha_min, min_idx = argmin(alpha[:, :n_h_set], axis=0)
                    min_idx = min_idx.tolist()[0]
                    alpha[:, :n_h_set] = repmat(alpha_min, l_var, 1)
                    D[:, h_set] = D[:, h_set] - multiply(
                        alpha[:, :n_h_set], D[:, h_set] - K[:, h_set])
                    D[min_idx, h_set] = 0
                    p_set[min_idx, h_set] = 0
                    K[:, h_set] = self.__cssls(
                        CtC, CtA[:, h_set], p_set[:, h_set])
                    h_set = find(any(K < 0, axis=0))
                    n_h_set = len(h_set)
            # make sure the solution has converged and check solution for
            # optimality
            W[:, f_set] = CtA[:, f_set] - dot(CtC, K[:, f_set])
            npw = multiply(np.logical_not(p_set[:, f_set]), W[:, f_set])
            j_set = find(all(npw <= 0, axis=0))
            f_j = f_set[j_set] if j_set != [] else []
            f_set = np.setdiff1d(np.asarray(f_set), np.asarray(f_j))
            # for non-optimal solutions, add the appropriate variable to Pset
            if len(f_set) > 0:
                _, mxidx = argmax(
                    multiply(np.logical_not(p_set[:, f_set]), W[:, f_set]), axis=0)
                mxidx = mxidx.tolist()[0]
                p_set[mxidx, f_set] = 1
                D[:, f_set] = K[:, f_set]
        return K

    def __cssls(self, CtC, CtA, p_set=None):
        """
        Solver for dense matrices. 
        
        Solve the set of equations CtA = CtC * K for variables defined in set p_set
        using the fast combinatorial approach (van Benthem and Keenan, 2004).
        """
        K = np.mat(np.zeros(CtA.shape))
        if p_set is None or p_set.size == 0 or all(p_set):
            # equivalent if CtC is square matrix
            K = np.linalg.lstsq(CtC, CtA, rcond=-1)[0]
            # K = dot(np.linalg.pinv(CtC), CtA)
        else:
            l_var, p_rhs = p_set.shape
            coded_p_set = dot(
                np.mat(2 ** np.array(list(range(l_var - 1, -1, -1)))), p_set)
            sorted_p_set, sorted_idx_set = sort(coded_p_set)
            breaks = diff(np.mat(sorted_p_set))
            break_idx = [-1] + find(np.mat(breaks)) + [p_rhs]
            for k in range(len(break_idx) - 1):
                cols2solve = sorted_idx_set[
                    break_idx[k] + 1: break_idx[k + 1] + 1]
                vars = p_set[:, sorted_idx_set[break_idx[k] + 1]]
                vars = [i for i in range(vars.shape[0]) if vars[i, 0]]
                if vars != [] and cols2solve != []:
                    A = CtC[:, vars][vars, :]
                    B = CtA[:, cols2solve][vars,:]
                    sol = np.linalg.lstsq(A, B, rcond=-1)[0]
                    i = 0
                    for c in cols2solve:
                        j = 0
                        for v in vars:
                            K[v, c] = sol[j, i]
                            j += 1
                        i += 1
                    # K[vars, cols2solve] = dot(np.linalg.pinv(CtC[vars, vars]),
                    # CtA[vars, cols2solve])
        return K

    def __str__(self):
        return self.name + " - " + self.version

    def __repr__(self):
        return self.name




"""
###################################
Nmf (``methods.factorization.nmf``)
###################################

**Standard Nonnegative Matrix Factorization (NMF)** [Lee2001]_, [Lee1999]_.

Based on Kullback-Leibler divergence, it uses simple multiplicative updates
[Lee2001]_, [Lee1999]_, enhanced to avoid numerical underflow [Brunet2004]_.
Based on Euclidean distance, it uses simple multiplicative updates [Lee2001]_.
Different objective functions can be used, namely Euclidean distance, divergence
or connectivity matrix convergence.

Together with a novel model selection mechanism, NMF is an efficient method for
identification of distinct molecular patterns and provides a powerful method for
class discovery. It appears to have higher resolution such as HC or SOM and to
be less sensitive to a priori selection of genes. Rather than separating gene
clusters based on distance computation, NMF detects context-dependent patterns
of gene expression in complex biological systems.

Besides usages in bioinformatics NMF can be applied to text analysis,
image processing, multiway clustering, environmetrics etc.

.. literalinclude:: /code/snippet_nmf_fro.py

.. literalinclude:: /code/snippet_nmf_div.py

.. literalinclude:: /code/snippet_nmf_conn.py

"""

from nimfa.models import *
from nimfa.utils import *
from nimfa.utils.linalg import *

__all__ = ['Nmf']


class Nmf(nmf_std.Nmf_std):
    """
    :param V: The target matrix to estimate.
    :type V: Instance of the :class:`scipy.sparse` sparse matrices types,
       :class:`numpy.ndarray`, :class:`numpy.matrix` or tuple of instances of
       the latter classes.

    :param seed: Specify method to seed the computation of a factorization. If
       specified :param:`W` and :param:`H` seeding must be None. If neither seeding
       method or initial fixed factorization is specified, random initialization is
       used.
    :type seed: `str` naming the method or :class:`methods.seeding.nndsvd.Nndsvd`
       or None

    :param W: Specify initial factorization of basis matrix W. Default is None.
       When specified, :param:`seed` must be None.
    :type W: :class:`scipy.sparse` or :class:`numpy.ndarray` or
       :class:`numpy.matrix` or None

    :param H: Specify initial factorization of mixture matrix H. Default is None.
       When specified, :param:`seed` must be None.
    :type H: Instance of the :class:`scipy.sparse` sparse matrices types,
       :class:`numpy.ndarray`, :class:`numpy.matrix`, tuple of instances of the
       latter classes or None

    :param rank: The factorization rank to achieve. Default is 30.
    :type rank: `int`

    :param n_run: It specifies the number of runs of the algorithm. Default is
       1. If multiple runs are performed, fitted factorization model with the
       lowest objective function value is retained.
    :type n_run: `int`

    :param callback: Pass a callback function that is called after each run when
       performing multiple runs. This is useful if one wants to save summary
       measures or process the result before it gets discarded. The callback
       function is called with only one argument :class:`models.mf_fit.Mf_fit` that
       contains the fitted model. Default is None.
    :type callback: `function`

    :param callback_init: Pass a callback function that is called after each
       initialization of the matrix factors. In case of multiple runs the function
       is called before each run (more precisely after initialization and before
       the factorization of each run). In case of single run, the passed callback
       function is called after the only initialization of the matrix factors.
       This is useful if one wants to obtain the initialized matrix factors for
       further analysis or additional info about initialized factorization model.
       The callback function is called with only one argument
       :class:`models.mf_fit.Mf_fit` that (among others) contains also initialized
       matrix factors. Default is None.
    :type callback_init: `function`

    :param track_factor: When :param:`track_factor` is specified, the fitted
        factorization model is tracked during multiple runs of the algorithm. This
        option is taken into account only when multiple runs are executed
        (:param:`n_run` > 1). From each run of the factorization all matrix factors
        are retained, which can be very space consuming. If space is the problem
        setting the callback function with :param:`callback` is advised which is
        executed after each run. Tracking is useful for performing some quality or
        performance measures (e.g. cophenetic correlation, consensus matrix,
        dispersion). By default fitted model is not tracked.
    :type track_factor: `bool`

    :param track_error: Tracking the residuals error. Only the residuals from
        each iteration of the factorization are retained. Error tracking is not
        space consuming. By default residuals are not tracked and only the final
        residuals are saved. It can be used for plotting the trajectory of the
        residuals.
    :type track_error: `bool`

    :param update: Type of update equations used in factorization. When specifying
       model parameter ``update`` can be assigned to:

           #. 'euclidean' for classic Euclidean distance update
              equations,
           #. 'divergence' for divergence update equations.
       By default Euclidean update equations are used.
    :type update: `str`

    :param objective: Type of objective function used in factorization. When
       specifying model parameter :param:`objective` can be assigned to:

            #. 'fro' for standard Frobenius distance cost function,
            #. 'div' for divergence of target matrix from NMF
               estimate cost function (KL),
            #. 'conn' for measuring the number of consecutive
               iterations in which the connectivity matrix has not
               changed.
       By default the standard Frobenius distance cost function is used.
    :type objective: `str`

    :param conn_change: Stopping criteria used only if for :param:`objective`
       function connectivity matrix measure is selected. It specifies the minimum
       required of consecutive iterations in which the connectivity matrix has not
       changed. Default value is 30.
    :type conn_change: `int`

    **Stopping criterion**

    Factorization terminates if any of specified criteria is satisfied.

    :param max_iter: Maximum number of factorization iterations. Note that the
       number of iterations depends on the speed of method convergence. Default
       is 30.
    :type max_iter: `int`

    :param min_residuals: Minimal required improvement of the residuals from the
       previous iteration. They are computed between the target matrix and its MF
       estimate using the objective function associated to the MF algorithm.
       Default is None.
    :type min_residuals: `float`

    :param test_conv: It indicates how often convergence test is done. By
       default convergence is tested each iteration.
    :type test_conv: `int`
    """
    def __init__(self, V, seed=None, W=None, H=None, rank=30, max_iter=30,
                 min_residuals=1e-5, test_conv=None, n_run=1, callback=None,
                 callback_init=None, track_factor=False, track_error=False,
                 update='euclidean', objective='fro', conn_change=30, **options):
        self.name = "nmf"
        self.aseeds = ["random", "fixed", "nndsvd", "random_c", "random_vcol"]
        nmf_std.Nmf_std.__init__(self, vars())
        if 'conn' not in self.objective:
            self.conn_change = None
        self._conn_change = 0
        self.update = getattr(self, self.update)
        self.objective = getattr(self, self.objective)
        self.tracker = mf_track.Mf_track() if self.track_factor and self.n_run > 1 \
                                              or self.track_error else None

    def factorize(self):
        """
        Compute matrix factorization.
         
        Return fitted factorization model.
        """
        for run in range(self.n_run):
            self.W, self.H = self.seed.initialize(
                self.V, self.rank, self.options)
            p_obj = c_obj = sys.float_info.max
            best_obj = c_obj if run == 0 else best_obj
            iter = 0
            if self.callback_init:
                self.final_obj = c_obj
                self.n_iter = iter
                mffit = mf_fit.Mf_fit(self)
                self.callback_init(mffit)
            while self.is_satisfied(p_obj, c_obj, iter):
                p_obj = c_obj if not self.test_conv or iter % self.test_conv == 0 else p_obj
                self.update()
                self._adjustment()
                iter += 1
                if not self.test_conv or iter % self.test_conv == 0:
                    c_obj = self.objective()
                if self.track_error:
                    self.tracker.track_error(run, c_obj)
            if self.callback:
                self.final_obj = c_obj
                self.n_iter = iter
                mffit = mf_fit.Mf_fit(self)
                self.callback(mffit)
            if self.track_factor:
                self.tracker.track_factor(
                    run, W=self.W, H=self.H, final_obj=c_obj, n_iter=iter)
            # if multiple runs are performed, fitted factorization model with
            # the lowest objective function value is retained
            if c_obj <= best_obj or run == 0:
                best_obj = c_obj
                self.n_iter = iter
                self.final_obj = c_obj
                mffit = mf_fit.Mf_fit(copy.deepcopy(self))

        mffit.fit.tracker = self.tracker
        return mffit

    def is_satisfied(self, p_obj, c_obj, iter):
        """
        Compute the satisfiability of the stopping criteria based on stopping
        parameters and objective function value.
        
        Return logical value denoting factorization continuation. 
        
        :param p_obj: Objective function value from previous iteration. 
        :type p_obj: `float`
        :param c_obj: Current objective function value.
        :type c_obj: `float`
        :param iter: Current iteration number. 
        :type iter: `int`
        """
        if self.max_iter and self.max_iter <= iter:
            return False
        if self.test_conv and iter % self.test_conv != 0:
            return True
        if self.conn_change is not None:
            return self.__is_satisfied(p_obj, c_obj, iter)
        if self.min_residuals and iter > 0 and p_obj - c_obj < self.min_residuals:
            return False
        if iter > 0 and c_obj > p_obj:
            return False
        return True

    def __is_satisfied(self, p_obj, c_obj, iter):
        """
        Compute the satisfiability of the stopping criteria if change of
        connectivity matrices is used for objective function.
        
        Return logical value denoting factorization continuation.   
        
        :param p_obj: Objective function value from previous iteration. 
        :type p_obj: `float`
        :param c_obj: Current objective function value.
        :type c_obj: `float`
        :param iter: Current iteration number. 
        :type iter: `int`
        """
        self._conn_change = 0 if c_obj == 1 else self._conn_change + 1
        if self._conn_change >= self.conn_change:
            return False
        return True

    def _adjustment(self):
        """Adjust small values to factors to avoid numerical underflow."""
        self.H = max(self.H, np.finfo(self.H.dtype).eps)
        self.W = max(self.W, np.finfo(self.W.dtype).eps)

    def euclidean(self):
        """Update basis and mixture matrix based on Euclidean distance multiplicative update rules."""
        self.H = multiply(
            self.H, elop(dot(self.W.T, self.V), dot(self.W.T, dot(self.W, self.H)), div))
        self.W = multiply(
            self.W, elop(dot(self.V, self.H.T), dot(self.W, dot(self.H, self.H.T)), div))

    def divergence(self):
        """Update basis and mixture matrix based on divergence multiplicative update rules."""
        H1 = repmat(self.W.sum(0).T, 1, self.V.shape[1])
        self.H = multiply(
            self.H, elop(dot(self.W.T, elop(self.V, dot(self.W, self.H), div)), H1, div))
        W1 = repmat(self.H.sum(1).T, self.V.shape[0], 1)
        self.W = multiply(
            self.W, elop(dot(elop(self.V, dot(self.W, self.H), div), self.H.T), W1, div))

    def fro(self):
        """Compute squared Frobenius norm of a target matrix and its NMF estimate."""
        R = self.V - dot(self.W, self.H)
        return multiply(R, R).sum()

    def div(self):
        """Compute divergence of target matrix from its NMF estimate."""
        Va = dot(self.W, self.H)
        return (multiply(self.V, sop(elop(self.V, Va, div), op=np.log)) - self.V + Va).sum()

    def conn(self):
        """
        Compute connectivity matrix and compare it to connectivity matrix
        from previous iteration.

        Return logical value denoting whether connectivity matrix has changed
        from previous iteration.
        """
        _, idx = argmax(self.H, axis=0)
        mat1 = repmat(idx, self.V.shape[1], 1)
        mat2 = repmat(idx.T, 1, self.V.shape[1])
        cons = elop(mat1, mat2, eq)
        if not hasattr(self, 'consold'):
            self.cons = cons
            self.consold = np.mat(np.logical_not(cons))
        else:
            self.consold = self.cons
            self.cons = cons
        conn_change = elop(self.cons, self.consold, ne).sum()
        return conn_change > 0

    def __str__(self):
        return '%s - update: %s obj: %s' % (self.name, self.update.__name__, self.objective.__name__)

    def __repr__(self):
        return self.name




"""
###################################
Random (``methods.seeding.random``)
###################################

Random is the simplest MF initialization method.

The entries of factors are drawn from a uniform distribution over
[0, max(target matrix)). Generated matrix factors are sparse matrices with the
default density parameter of 0.01.
"""

from nimfa.utils.linalg import *

__all__ = ['Random']

class Random(object):

    def __init__(self):
        self.name = "random"

    def initialize(self, V, rank, options):
        """
        Return initialized basis and mixture matrix (and additional factors if
        specified in :param:`Sn`, n = 1, 2, ..., k).
        Initialized matrices are of the same type as passed target matrix. 
        
        :param V: Target matrix, the matrix for MF method to estimate.
        :type V: One of the :class:`scipy.sparse` sparse matrices types or
                :class:`numpy.matrix`
        :param rank: Factorization rank. 
        :type rank: `int`
        :param options: Specify the algorithm and model specific options (e.g. initialization of
                extra matrix factor, seeding parameters).
                    
                Option ``Sn``, n = 1, 2, 3, ..., k specifies additional k matrix factors which
                need to be initialized. The value of each option Sn is a tuple denoting matrix
                shape. Matrix factors are returned in the same order as their descriptions in input.

                Option ``density`` represents density of generated matrices. Density of 1 means a
                full matrix, density of 0 means a matrix with no nonzero items. Default value is 0.7.
                Density parameter is applied only if passed target ``V`` is an instance of one :class:`scipy.sparse` sparse types.
        :type options: `dict`
        """
        self.rank = rank
        self.density = options.get('density', 0.7)
        if sp.isspmatrix(V):
            self.max = V.data.max()
            self._format = V.getformat()
            gen = self.gen_sparse
        else:
            self.max = V.max()
            self.prng = np.random.RandomState()
            gen = self.gen_dense
        self.W = gen(V.shape[0], self.rank)
        self.H = gen(self.rank, V.shape[1])
        mfs = [self.W, self.H]
        for sn in options:
            if sn[0] is 'S' and sn[1:].isdigit():
                mfs.append(gen(options[sn][0], options[sn][1]))
        return mfs

    def gen_sparse(self, dim1, dim2):
        """
        Return randomly initialized sparse matrix of specified dimensions.
        
        :param dim1: Dimension along first axis.
        :type dim1: `int`
        :param dim2: Dimension along second axis.
        :type dim2: `int`
        """
        rnd = sp.rand(dim1, dim2, density=self.density, format=self._format)
        return abs(self.max * rnd)

    def gen_dense(self, dim1, dim2):
        """
        Return randomly initialized :class:`numpy.matrix` matrix of specified
        dimensions.
        
        :param dim1: Dimension along first axis.
        :type dim1: `int`
        :param dim2: Dimension along second axis.
        :type dim2: `int`
        """
        return np.mat(self.prng.uniform(0, self.max, (dim1, dim2)))

    def __repr__(self):
        return "random.Random()"

    def __str__(self):
        return self.name




"""
#######################################
Random_c (``methods.seeding.random_c``)
#######################################

Random C [Albright2006]_ is inexpensive initialization method for nonnegative
matrix factorization. It is inspired by the C matrix in of the CUR decomposition.
The Random C initialization is similar to the Random Vcol method (see
mod:`methods.seeding.random_vcol`) except it chooses p columns at random from
the longest (in 2-norm) columns in target matrix (V), which generally means the
most dense columns of target matrix.

Initialization of each column of basis matrix is done by averaging p random
columns of l longest columns of target matrix. Initialization
of mixture matrix is similar except for row operations.    
"""

from nimfa.utils.linalg import *

__all__ = ['Random_c']


class Random_c(object):

    def __init__(self):
        self.name = "random_c"

    def initialize(self, V, rank, options):
        """
        Return initialized basis and mixture matrix. Initialized matrices are
        of the same type as passed target matrix.
        
        :param V: Target matrix, the matrix for MF method to estimate. 
        :type V: One of the :class:`scipy.sparse` sparse matrices types or or
                :class:`numpy.matrix`
        :param rank: Factorization rank. 
        :type rank: `int`
        :param options: Specify the algorithm and model specific options (e.g. initialization of
                extra matrix factor, seeding parameters).

                Option ``p_c`` represents the number of columns of target matrix
                used to average the column of basis matrix. Default
                value for ``p_c`` is 1/5 * (target.shape[1]).

                Option ``p_r`` represents the number of rows of target matrix used
                to average the row of basis matrix. Default value for
                ``p_r`` is 1/5 * (target.shape[0]).

                Option ``l_c`` represents the first l_c columns of target matrix sorted
                descending by length (2-norm). Default value for ``l_c`` is  1/2 * (target.shape[1]).

                Option ``l_r`` represent first l_r rows of target matrix sorted
                descending by length (2-norm). Default value for
                ``l_r`` is 1/2 * (target.shape[0]).
        :type options: `dict`
        """
        self.rank = rank
        self.p_c = options.get('p_c', int(ceil(1. / 5 * V.shape[1])))
        self.p_r = options.get('p_r', int(ceil(1. / 5 * V.shape[0])))
        self.l_c = options.get('l_c', int(ceil(1. / 2 * V.shape[1])))
        self.l_r = options.get('l_r', int(ceil(1. / 2 * V.shape[0])))
        self.prng = np.random.RandomState()
        if sp.isspmatrix(V):
            self.W = sp.lil_matrix((V.shape[0], self.rank))
            self.H = sp.lil_matrix((self.rank, V.shape[1]))
            top_c = sorted(enumerate([norm(V[:, i], 2)
                           for i in range(V.shape[1])]), key=itemgetter(1), reverse=True)[:self.l_c]
            top_r = sorted(
                enumerate([norm(V[i, :], 2) for i in range(V.shape[0])]), key=itemgetter(1), reverse=True)[:self.l_r]
        else:
            self.W = np.mat(np.zeros((V.shape[0], self.rank)))
            self.H = np.mat(np.zeros((self.rank, V.shape[1])))
            top_c = sorted(enumerate([norm(V[:, i], 2)
                           for i in range(V.shape[1])]), key=itemgetter(1), reverse=True)[:self.l_c]
            top_r = sorted(
                enumerate([norm(V[i, :], 2) for i in range(V.shape[0])]), key=itemgetter(1), reverse=True)[:self.l_r]
        top_c = np.mat(list(zip(*top_c))[0])
        top_r = np.mat(list(zip(*top_r))[0])
        for i in range(self.rank):
            self.W[:, i] = V[
                :, top_c[0, self.prng.randint(low=0, high=self.l_c, size=self.p_c)].tolist()[0]].mean(axis=1)
            self.H[i, :] = V[
                top_r[0, self.prng.randint(low=0, high=self.l_r, size=self.p_r)].tolist()[0], :].mean(axis=0)
        # return sparse or dense initialization
        if sp.isspmatrix(V):
            return self.W.tocsr(), self.H.tocsr()
        else:
            return self.W, self.H

    def __repr__(self):
        return "random_c.Random_c()"

    def __str__(self):
        return self.name




"""    
    This package contains implementations of initialization methods for matrix factorization.
"""

from .nndsvd import *
from .random import *
from .fixed import *
from .random_c import *
from .random_vcol import *

methods = {"random": Random,
           "fixed": Fixed,
           "nndsvd": Nndsvd,
           "random_c": Random_c,
           "random_vcol": Random_vcol,
           "none": None}



import numpy as np

import nimfa

V = np.random.rand(30, 20)

init_W = np.random.rand(30, 4)
init_H = np.random.rand(4, 20)

# Fixed initialization of latent matrices
nmf = nimfa.Nmf(V, seed="fixed", W=init_W, H=init_H, rank=4)
nmf_fit = nmf()

print("Euclidean distance: %5.3f" % nmf_fit.distance(metric="euclidean"))
print('Initialization type: %s' % nmf_fit.seeding)
print('Iterations: %d' % nmf_fit.n_iter)




"""
    ##########################
    Mf_fit (``models.mf_fit``)
    ##########################
"""


class Mf_fit():
    """
    Base class for storing MF results.
    
    It contains generic functions and structure for handling the results of MF algorithms. 
    It contains a slot with the fitted MF model and data about parameters and methods used for
    factorization.  
    
    The purpose of this class is to handle in a generic way the results of MF algorithms and acts as a wrapper for the 
    fitted model. Its attribute attribute:: fit contains the fitted model and its configuration 
    can therefore be used directly in following calls to factorization.    
    
    .. attribute:: fit
        
        The fitted NMF model
    
    .. attribute:: algorithm 

        NMF method of factorization.

    .. attribute:: n_iter

        The number of iterations performed.

    .. attribute:: n_run

        The number of NMF runs performed.

    .. attribute:: seeding

        The seeding method used to seed the algorithm that fitted NMF model. 

    .. attribute:: options

        Extra parameters specific to the algorithm used to fit the model.
    """
    def __init__(self, fit):
        """
        Construct fitted factorization model. 
        
        :param fit: Matrix factorization algorithm model. 
        :type fit: class from methods.mf package
        """
        self.fit = fit
        self.algorithm = str(self.fit)
        self.n_iter = self.fit.n_iter
        self.n_run = self.fit.n_run
        self.seeding = str(self.fit.seed)
        self.options = self.fit.options

    def basis(self):
        """Return the matrix of basis vectors."""
        return self.fit.basis()

    def coef(self, idx=None):
        """
        Return the matrix of mixture coefficients.
        
        :param idx: Name of the matrix (coefficient) matrix. Used only in the multiple NMF model.
        :type idx: `str` with values 'coef' or 'coef1' (`int` value of 0 or 1, respectively) 
        """
        return self.fit.coef(idx)

    def distance(self, metric=None, idx=None):
        """
        Return the loss function value. If metric is not supplied, final objective function value associated to the MF algorithm is returned.
        
        :param metric: Measure of distance between a target matrix and a MF estimate. Metric 'kl' and 'euclidean' 
           are defined.
        :type metric: 'str'

        :param idx: Name of the matrix (coefficient) matrix. Used only in the multiple NMF model.
        :type idx: `str` with values 'coef' or 'coef1' (`int` value of 0 or 1, respectively) 
        """
        if metric == None:
            return self.fit.final_obj
        else:
            return self.fit.distance(metric, idx)

    def fitted(self, idx=None):
        """
        Compute the estimated target matrix according to the MF algorithm model.
        
        :param idx: Name of the matrix (coefficient) matrix. Used only in the multiple NMF model.
        :type idx: `str` with values 'coef' or 'coef1' (`int` value of 0 or 1, respectively) 
        """
        return self.fit.fitted(idx)

    def fit(self):
        """Return the MF algorithm model."""
        return self.fit

    def summary(self, idx=None):
        """
        Return generic set of measures to evaluate the quality of the factorization.
        
        :param idx: Name of the matrix (coefficient) matrix. Used only in the multiple NMF model.
        :type idx: `str` with values 'coef' or 'coef1' (`int` value of 0 or 1, respectively) 
        """
        if idx == 'coef':
            idx = 0
        if idx == 'coef1':
            idx = 1
        if hasattr(self, 'summary_data'):
            if idx not in self.summary_data:
                self.summary_data[idx] = self._compute_summary(idx)
            return self.summary_data[idx]
        else:
            self.summary_data = {}
            self.summary_data[idx] = self._compute_summary(idx)
            return self.summary_data[idx]

    def _compute_summary(self, idx=None):
        """
        Compute generic set of measures to evaluate the quality of the factorization.
        
        :param idx: Name of the matrix (coefficient) matrix. Used only in the multiple NMF model.
        :type idx: `str` with values 'coef' or 'coef1' (`int` value of 0 or 1, respectively) 
        """
        return {
            'rank': self.fit.rank,
            'sparseness': self.fit.sparseness(idx=idx),
            'rss': self.fit.rss(idx=idx),
            'evar': self.fit.evar(idx=idx),
            'residuals': self.fit.residuals(idx=idx),
            'connectivity': self.fit.connectivity(idx=idx),
            'predict_samples': self.fit.predict(what='samples', prob=True, idx=idx),
            'predict_features': self.fit.predict(what='features', prob=True, idx=idx),
            'score_features': self.fit.score_features(idx=idx),
            'select_features': self.fit.select_features(idx=idx),
            'dispersion': self.fit.dispersion(idx=idx),
            'cophenetic': self.fit.coph_cor(idx=idx),
            'consensus': self.fit.consensus(idx=idx),
            'euclidean': self.fit.distance(metric='euclidean', idx=idx),
            'kl': self.fit.distance(metric='kl', idx=idx),
            'n_iter': self.fit.n_iter,
            'n_run': self.fit.n_run
        }




"""
###################################
Bmf (``methods.factorization.bmf``)
###################################

**Binary Matrix Factorization (BMF)** [Zhang2007]_.

BMF extends standard NMF to binary matrices. Given a binary target matrix (V),
we want to factorize it into binary basis and mixture matrices, thus conserving
the most important integer property of the target matrix. Common methodologies
include penalty function algorithm and thresholding algorithm. 

BMF can be derived based on variant of Standard NMF, but some problems need to
be resolved:
    
    #. Uniqueness. Solution for basis and mixture matrix is not unique as it is
       always possible to find a diagonal matrix and incorporate it current solution
       to get a new.
    #. Scale. Scale problem arises when discretizing basis and mixture matrix into
       binary matrices. This problem can be resolved by using rescaling proposed in
       Boundedness Theorem in [Zhang2007]_. Therefore, discretization works properly
       because basis and mixture matrix are in the same scales. The factorization
       method is more robust in this way. It has been shown that the percentage of
       nonzero elements in normalized case is lower than in nonnormalized case.
       Without normalization the mixture matrix is often very sparse and the basis
       matrix very dense - much information, given via mixture matrix is lost and
       cannot be compensated with basis matrix.

This method implements penalty function algorithm. The problem of BMF can be
represented in terms of nonlinear programming and then solved by a penalty
function algorithm. The algorithm is described as follows:

    1. Initialize basis, mixture matrix and parameters. 
    2. Normalize basis and mixture using Boundedness Theorem in [Zhang2007]_.
    3. For basis and mixture, alternately solve nonlinear optimization problem
       with the objective function composed of three components: Euclidean distance
       of BMF estimate from target matrix; mixture penalty term and  basis penalty
       term.
    4. Update parameters based on the level of the binarization of the basis and
       mixture matrix.
    
In step 1, basis and mixture matrix can be initialized with common initialization
methods or with the result of the Standard NMF by passing fixed factors to the
factorization model. In step 3, the update rule is derived by taking the longest
step that can maintain the nonnegativity of the basis, mixture matrix during the
iterative process.

.. literalinclude:: /code/snippet_bmf.py

"""

from nimfa.models import *
from nimfa.utils import *
from nimfa.utils.linalg import *

__all__ = ['Bmf']


class Bmf(nmf_std.Nmf_std):
    """
    :param V: The target matrix to estimate.
    :type V: Instance of the :class:`scipy.sparse` sparse matrices types,
       :class:`numpy.ndarray`, :class:`numpy.matrix` or tuple of instances of
       the latter classes.

    :param seed: Specify method to seed the computation of a factorization. If
       specified :param:`W` and :param:`H` seeding must be None. If neither seeding
       method or initial fixed factorization is specified, random initialization is
       used.
    :type seed: `str` naming the method or :class:`methods.seeding.nndsvd.Nndsvd`
       or None

    :param W: Specify initial factorization of basis matrix W. Default is None.
       When specified, :param:`seed` must be None.
    :type W: :class:`scipy.sparse` or :class:`numpy.ndarray` or
       :class:`numpy.matrix` or None

    :param H: Specify initial factorization of mixture matrix H. Default is None.
       When specified, :param:`seed` must be None.
    :type H: Instance of the :class:`scipy.sparse` sparse matrices types,
       :class:`numpy.ndarray`, :class:`numpy.matrix`, tuple of instances of the
       latter classes or None

    :param rank: The factorization rank to achieve. Default is 30.
    :type rank: `int`

    :param n_run: It specifies the number of runs of the algorithm. Default is
       1. If multiple runs are performed, fitted factorization model with the
       lowest objective function value is retained.
    :type n_run: `int`

    :param callback: Pass a callback function that is called after each run when
       performing multiple runs. This is useful if one wants to save summary
       measures or process the result before it gets discarded. The callback
       function is called with only one argument :class:`models.mf_fit.Mf_fit` that
       contains the fitted model. Default is None.
    :type callback: `function`

    :param callback_init: Pass a callback function that is called after each
       initialization of the matrix factors. In case of multiple runs the function
       is called before each run (more precisely after initialization and before
       the factorization of each run). In case of single run, the passed callback
       function is called after the only initialization of the matrix factors.
       This is useful if one wants to obtain the initialized matrix factors for
       further analysis or additional info about initialized factorization model.
       The callback function is called with only one argument
       :class:`models.mf_fit.Mf_fit` that (among others) contains also initialized
       matrix factors. Default is None.
    :type callback_init: `function`

    :param track_factor: When :param:`track_factor` is specified, the fitted
        factorization model is tracked during multiple runs of the algorithm. This
        option is taken into account only when multiple runs are executed
        (:param:`n_run` > 1). From each run of the factorization all matrix factors
        are retained, which can be very space consuming. If space is the problem
        setting the callback function with :param:`callback` is advised which is
        executed after each run. Tracking is useful for performing some quality or
        performance measures (e.g. cophenetic correlation, consensus matrix,
        dispersion). By default fitted model is not tracked.
    :type track_factor: `bool`

    :param track_error: Tracking the residuals error. Only the residuals from
        each iteration of the factorization are retained. Error tracking is not
        space consuming. By default residuals are not tracked and only the final
        residuals are saved. It can be used for plotting the trajectory of the
        residuals.
    :type track_error: `bool`

    :param lambda_w: It controls how fast lambda should increase and influences
       the convergence of the basis matrix (W) to binary values during the update.

             #. ``lambda_w`` < 1 will result in a nonbinary
                decompositions as the update rule effectively is a
                conventional NMF update rule.
             #. ``lambda_w`` > 1 give more weight to make the
                factorization binary with increasing iterations.
       Default value is 1.1.
    :type lambda_w: `float`

    :param lambda_h: It controls how fast lambda should increase and influences
       the convergence of the mixture matrix (H) to binary values during the update.

             #. ``lambda_h`` < 1 will result in a nonbinary
                decompositions as the update rule effectively is a
                conventional NMF update rule.
             #. ``lambda_h`` > 1 give more weight to make the
                factorization binary with increasing iterations.
       Default value is 1.1.
    :type lambda_h: `float`

    **Stopping criterion**

    Factorization terminates if any of specified criteria is satisfied.

    :param max_iter: Maximum number of factorization iterations. Note that the
       number of iterations depends on the speed of method convergence. Default
       is 30.
    :type max_iter: `int`

    :param min_residuals: Minimal required improvement of the residuals from the
       previous iteration. They are computed between the target matrix and its MF
       estimate using the objective function associated to the MF algorithm.
       Default is None.
    :type min_residuals: `float`

    :param test_conv: It indicates how often convergence test is done. By
       default convergence is tested each iteration.
    :type test_conv: `int`
    """
    def __init__(self, V, seed=None, W=None, H=None, rank=30, max_iter=30,
                 min_residuals=1e-5, test_conv=None, n_run=1, callback=None,
                 callback_init=None, track_factor=False, track_error=False,
                 lambda_w=1.1, lambda_h=1.1, **options):
        self.name = "bmf"
        self.aseeds = ["random", "fixed", "nndsvd", "random_c", "random_vcol"]
        nmf_std.Nmf_std.__init__(self, vars())
        self.tracker = mf_track.Mf_track() if self.track_factor and self.n_run > 1 \
                                              or self.track_error else None

    def factorize(self):
        """
        Compute matrix factorization.
         
        Return fitted factorization model.
        """
        self._lambda_w = 1. / self.max_iter if self.max_iter else 1. / 10
        self._lambda_h = self._lambda_w
        for run in range(self.n_run):
            self.W, self.H = self.seed.initialize(
                self.V, self.rank, self.options)
            self.normalize()
            p_obj = c_obj = sys.float_info.max
            best_obj = c_obj if run == 0 else best_obj
            iter = 0
            if self.callback_init:
                self.final_obj = c_obj
                self.n_iter = iter
                mffit = mf_fit.Mf_fit(self)
                self.callback_init(mffit)
            while self.is_satisfied(p_obj, c_obj, iter):
                p_obj = c_obj if not self.test_conv or iter % self.test_conv == 0 else p_obj
                self.update()
                self._adjustment()
                iter += 1
                c_obj = self.objective(
                ) if not self.test_conv or iter % self.test_conv == 0 else c_obj
                if self.track_error:
                    self.tracker.track_error(run, c_obj)
            if self.callback:
                self.final_obj = c_obj
                self.n_iter = iter
                mffit = mf_fit.Mf_fit(self)
                self.callback(mffit)
            if self.track_factor:
                self.tracker.track_factor(
                    run, W=self.W, H=self.H, final_obj=c_obj, n_iter=iter)
            # if multiple runs are performed, fitted factorization model with
            # the lowest objective function value is retained
            if c_obj <= best_obj or run == 0:
                best_obj = c_obj
                self.n_iter = iter
                self.final_obj = c_obj
                mffit = mf_fit.Mf_fit(copy.deepcopy(self))

        mffit.fit.tracker = self.tracker
        return mffit

    def is_satisfied(self, p_obj, c_obj, iter):
        """
        Compute the satisfiability of the stopping criteria based on stopping
        parameters and objective function value.
        
        Return logical value denoting factorization continuation. 
        
        :param p_obj: Objective function value from previous iteration. 
        :type p_obj: `float`
        :param c_obj: Current objective function value.
        :type c_obj: `float`
        :param iter: Current iteration number. 
        :type iter: `int`
        """
        if self.max_iter and self.max_iter <= iter:
            return False
        if self.test_conv and iter % self.test_conv != 0:
            return True
        if self.min_residuals and iter > 0 and p_obj - c_obj < self.min_residuals:
            return False
        if iter > 0 and c_obj > p_obj:
            return False
        return True

    def update(self):
        """Update basis and mixture matrix."""
        # update mixture matrix
        H1 = dot(self.W.T, self.V) + 3. * \
            self._lambda_h * multiply(self.H, self.H)
        H2 = dot(dot(self.W.T, self.W), self.H) + 2. * \
            self._lambda_h * power(self.H, 3) + self._lambda_h * self.H
        self.H = multiply(self.H, elop(H1, H2, div))
        # update basis matrix,
        W1 = dot(self.V, self.H.T) + 3. * \
            self._lambda_w * multiply(self.W, self.W)
        W2 = dot(self.W, dot(self.H, self.H.T)) + 2. * \
            self._lambda_w * power(self.W, 3) + self._lambda_w * self.W
        self.W = multiply(self.W, elop(W1, W2, div))
        self._lambda_h = self.lambda_h * self._lambda_h
        self._lambda_w = self.lambda_w * self._lambda_w

    def normalize(self):
        """
        Normalize initialized basis and mixture matrix, using Boundedness
        Theorem in [Zhang2007]_. Normalization makes the BMF factorization more
        robust.
        
        Normalization produces basis and mixture matrix with values in [0, 1]. 
        """
        val_w, _ = argmax(self.W, axis=0)
        val_h, _ = argmax(self.H, axis=1)
        if sp.isspmatrix(self.W):
            D_w = sp.spdiags(val_w, 0, self.W.shape[1], self.W.shape[1])
        else:
            D_w = np.diag(val_w)
        if sp.isspmatrix(self.H):
            D_h = sp.spdiags(val_h, 0, self.H.shape[0], self.H.shape[0])
        else:
            D_h = np.diag(val_h)
        self.W = dot(dot(self.W, power(D_w, -0.5)), power(D_h, 0.5))
        self.H = dot(dot(power(D_h, -0.5), power(D_w, 0.5)), self.H)

    def objective(self):
        """Compute squared Frobenius norm of a target matrix and its NMF estimate."""
        R = self.V - dot(self.W, self.H)
        return (multiply(R, R)).sum()

    def _adjustment(self):
        """Adjust small values to factors to avoid numerical underflow."""
        self.H = max(self.H, np.finfo(self.H.dtype).eps)
        self.W = max(self.W, np.finfo(self.W.dtype).eps)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name
